module TestRumkin exposing (suiteGetCharsetSize, suiteParseCommon, suiteParseFrequency, suiteRumkin)

import Array
import Expect exposing (FloatingPointTolerance(..))
import Internal.Rumkin exposing (CommonPasswordList(..), FrequencyList(..), getCharsetSize)
import Rumkin exposing (Strength(..), getStats, parseCommonList, parseFrequencyList)
import Rumkin.Common exposing (commonList)
import Rumkin.Frequency exposing (frequencyList)
import Set
import Test exposing (Test, test)


suiteRumkin : Test
suiteRumkin =
    test "All passwords should yield correct results." <|
        \() ->
            let
                cases =
                    [ ( "asds", { entropy = 13.2, strength = VeryWeak, warnings = [ "Very short password!" ] } )
                    , ( "02/13/1960", { entropy = 26.2, strength = VeryWeak, warnings = [] } )
                    , ( "EpyZFzkLNy2TEe4jd6hsV9", { entropy = 112.8, strength = Strong, warnings = [] } )
                    , ( "C,;z.#7dp<E3ah6S+4N,aHzn\\@", { entropy = 130.4, strength = VeryStrong, warnings = [] } )
                    , ( "MûËiÐº7JÏ2,2±DVå7®yÐtf", { entropy = 124.5, strength = Strong, warnings = [] } )
                    , ( "passwordlookhere", { entropy = 60.4, strength = Strong, warnings = [ "Common password!" ] } )

                    -- Make sure unicode treatment is similar
                    , ( "隙生人", { entropy = 8.5, strength = VeryWeak, warnings = [ "Very short password!" ] } )
                    ]

                results =
                    List.concatMap
                        (\( p, e ) ->
                            let
                                r =
                                    getStats p
                            in
                            [ \() -> Expect.within (Absolute 0.000000001) e.entropy r.entropy
                            , \() -> Expect.equal e.strength r.strength
                            , \() -> Expect.equal e.warnings r.warnings
                            ]
                        )
                        cases
            in
            Expect.all results ()


suiteGetCharsetSize : Test
suiteGetCharsetSize =
    test "All charsets should be calculated accurately" <|
        \() ->
            let
                examples =
                    [ ( "ngwnhuxjkxmtsqqnvvyuww", 26 )
                    , ( "SLCWAADNUTKRPUTVRXUCGH", 26 )
                    , ( "2947337354874496527482", 10 )
                    , ( "!#$%&*", 10 )
                    , ( "+[{]", 22 )
                    , ( ".,!'{+\"^&--\\)$#\\~_{?%#", 32 )
                    , ( "äÈÈ¦ÅÊÆÐ²êëÃÛ×¢þÓ·£ìÔº", 160 )
                    , ( "×)¥^$^.êì÷é^ÉèÄó¡=»%^Ç", 192 )
                    , ( "MûËiÐº7JÏ2,2±DVå7®yÐtf", 244 )

                    -- Make certain unicode treatment is the same
                    , ( "隙生人", 160 )
                    ]

                results =
                    List.map (\( p, e ) -> \() -> Expect.equal e <| getCharsetSize p) examples
            in
            Expect.all results ()


suiteParseFrequency : Test
suiteParseFrequency =
    test "Confirm parsing frequency list matches reference." <|
        \() ->
            let
                list =
                    parseFrequencyList frequencyList
                        |> Result.withDefault (FrequencyList Array.empty)
                        |> (\f ->
                                case f of
                                    FrequencyList a ->
                                        Array.toList a
                           )
            in
            Expect.all
                (List.map2 (\f1 f2 -> \() -> Expect.within (Absolute 0.000000001) f1 f2)
                    [ 0.23653710453418866, 0.04577693541332556, 0.03449832337075375, 0.042918209651552706, 0.037390873305146524, 0.028509112115468728, 0.02350896632162123, 0.022188657238664526, 0.028429800262428927, 0.04357019973757107, 0.00913602565971716, 0.03223093745443942, 0.02235311269864412, 0.04438081352966905, 0.04512377897652719, 0.020055401662049863, 0.055903192885260244, 0.0024388394809739026, 0.035207464644991984, 0.07355941099285611, 0.036905671380667734, 0.026134421927394666, 0.023787724158040528, 0.011352092141711621, 0.0032354570637119114, 0.005986878553725033, 0.008861933226417843, 0.11511532293337222, 0.027556203528211108, 0.024331243621519172, 0.039266365359381834, 0.031599941682461, 0.014403265782183991, 0.015480973902901297, 0.027770812071730572, 0.00942761335471643, 0.039872867764980315, 0.0078122175244204695, 0.02808456043154979, 0.08429100451960927, 0.04688963405744277, 0.13831170724595424, 0.002540311998833649, 0.025211838460416972, 0.001543082081936142, 0.09519638431258201, 0.061845750109345385, 0.08907071001603732, 0.02137571074500656, 0.027093162268552268, 0.005521504592506197, 0.003023181221752442, 0.007086747339262283, 0.010262720513194342, 0.08785070710016038, 0.14617757690625455, 0.03417291150313457, 0.0059635515381250915, 0.006146668610584633, 0.195202799241872, 0.002774748505613063, 0.004715556203528212, 0.0044776206444088066, 0.11205481848665985, 0.005654468581425864, 0.0028820527773727946, 0.07383000437381543, 0.005516839189386207, 0.006496573844583759, 0.09843067502551392, 0.0027140982650532145, 0.0006893133109782768, 0.08425368129464937, 0.021325557661466685, 0.006493074792243767, 0.07023414491908442, 0.002077270739174807, 0.0024633328473538415, 0.0007744569179180639, 0.015413325557661468, 0.0011990086018370024, 0.13162851727657093, 0.10115993585070711, 0.0026989357049132527, 0.03319317684793702, 0.002946202070272634, 0.0783216212275842, 0.0018358361277154103, 0.00258813238081353, 0.2141688292754046, 0.09853681294649366, 0.0032482869222918796, 0.04359352675317102, 0.01993526753171016, 0.0036880011663507797, 0.008011663507799971, 0.12014696019827964, 0.0029846916460125384, 0.0017553579238956116, 0.029470185158186325, 0.010413179763813967, 0.030699518880303252, 0.03508499781309229, 0.002021285901734947, 0.0010613792097973467, 0.0005295232541186761, 0.009677212421635807, 0.010585799679253535, 0.17101734946785244, 0.07968625164018078, 0.007839043592360402, 0.005438693687126403, 0.0183606939787141, 0.2732701559994168, 0.004953491762647616, 0.007259367254701851, 0.008104971570199739, 0.13274588132380813, 0.004210526315789474, 0.004997813092287506, 0.017006560723137484, 0.007442484327161393, 0.016789619478058026, 0.08477737279486806, 0.005106283714827234, 0.0005026971861787433, 0.04040355736987899, 0.037535500801866156, 0.00885960052485785, 0.0336410555474559, 0.007066919376002332, 0.005344219273946639, 0.0006333284735384167, 0.010684939495553289, 0.0063064586674442345, 0.15386849394955532, 0.015049424114302375, 0.012162705933809595, 0.020425134859308938, 0.037366379938766583, 0.02157165767604607, 0.009373961218836564, 0.0173214754337367, 0.009616562181075958, 0.029522670943286193, 0.010154249890654615, 0.018600962239393497, 0.06362210234728094, 0.03157078291296107, 0.151603440734801, 0.0062329785683044175, 0.014775331681003062, 0.0020854351946347867, 0.1826342032366234, 0.0878017203674005, 0.054190989940224525, 0.010329202507654177, 0.012763376585508092, 0.0064872430383437815, 0.006381105117364048, 0.005388540603586529, 0.0090800408222773, 0.09611196967487973, 0.09940691062837148, 0.01033969966467415, 0.004034407348009914, 0.008826942703017933, 0.11474675608689314, 0.07132584924916169, 0.012388977985129028, 0.005435194634786413, 0.1417174515235457, 0.0037066627788307337, 0.0045802595130485495, 0.060800699810468, 0.005341886572386646, 0.005683627350925791, 0.12434932205860913, 0.004596588423968508, 0.0007534626038781163, 0.07107041842834232, 0.022361277154104096, 0.04784720804782038, 0.06277533168100306, 0.003441901151771395, 0.005828254847645429, 0.0009669047966175828, 0.009470768333576322, 0.002077270739174807, 0.12797667298440007, 0.08797783933518005, 0.005388540603586529, 0.0024913252660737715, 0.007550954949701123, 0.2786866890217233, 0.002509986878553725, 0.029002478495407494, 0.0303204548768042, 0.07576614666861058, 0.00246799825047383, 0.00592389561160519, 0.039574281965301064, 0.00706808572678233, 0.03304505029887739, 0.05474150750838315, 0.0028633911648928414, 0.0005073625892987316, 0.07293541332555767, 0.053528502697186175, 0.022566554891383584, 0.038151334013704616, 0.002716430966613209, 0.005049132526607377, 0.0009902318122175246, 0.008997229916897508, 0.0011861787432570347, 0.1666377022889634, 0.14414462749671964, 0.003374252806531564, 0.005169266656947077, 0.008468873013558828, 0.16337541915731155, 0.002873888321912815, 0.004305000728969237, 0.0031141565825922144, 0.1241172182533897, 0.0052800699810468, 0.008969237498177577, 0.024094474413179766, 0.017029887738737422, 0.01722700102055693, 0.10618457501093455, 0.006147834961364631, 0.0008269427030179326, 0.03303571949263741, 0.024188948826359528, 0.05213937891820965, 0.04505846333284735, 0.0035270447587111824, 0.006799825047383001, 0.0008199445983379502, 0.02206735675754483, 0.001010059775477475, 0.11971191135734072, 0.04656538854060359, 0.011243621519171892, 0.06513019390581717, 0.032375564951159064, 0.06347047674588133, 0.013678961947805804, 0.03309870243475726, 0.006982942119842543, 0.009726199154395685, 0.010121592068814697, 0.032514360693978714, 0.04986032949409535, 0.039734072022160664, 0.15690683773144773, 0.03949963551538125, 0.014790494241143023, 0.002722262720513194, 0.02614375273363464, 0.10753637556495116, 0.06764834523983088, 0.006221315060504448, 0.021317393206006705, 0.0030826651115322934, 0.002399183554454002, 0.0019069835252952323, 0.015595276279341012, 0.0925126111678087, 0.18437906400349907, 0.006538562472663654, 0.008719638431258201, 0.02116693395538708, 0.18241376293920394, 0.007290858725761773, 0.005976381396705059, 0.005629975215045925, 0.09721300481119698, 0.004810030616707975, 0.024303251202799244, 0.012954658113427612, 0.011057005394372358, 0.02733459688001166, 0.10135121737862662, 0.012016912086309959, 0.001055547455897361, 0.009027555037177431, 0.07162326869806095, 0.01007143898527482, 0.07297623560285756, 0.006741507508383147, 0.0036891675171307776, 0.0008409389123778977, 0.011272780288671819, 0.007020265344802449, 0.1030389269572824, 0.15350809155853623, 0.004232686980609419, 0.004353987461729115, 0.0023385333138941536, 0.14450386353695874, 0.002546143752733635, 0.0024470039364338824, 0.01200758128006998, 0.0981227584195947, 0.003161976964572095, 0.040695145064878264, 0.03460446129173349, 0.003908441463770229, 0.01598483743986004, 0.13107216795451232, 0.003129319142732177, 0.00032307916605919226, 0.04050386353695874, 0.05452689896486368, 0.03589677795597026, 0.07087097244496282, 0.006143169558244642, 0.008684647907858289, 0.0004607085580988482, 0.022010205569324977, 0.0009097536083977258, 0.07328765126111678, 0.14751421490013122, 0.008015162560139961, 0.006601545414783497, 0.025279486805656802, 0.1682449336637994, 0.008313748359819215, 0.007010934538562473, 0.005886572386645284, 0.16889575739903775, 0.004123050007289692, 0.011925936725470185, 0.10007289692374982, 0.013380376148126549, 0.009021723283277445, 0.08650823735238372, 0.007756232686980609, 0.0007243038343781893, 0.0026791077416533026, 0.02797492345823006, 0.032384895757399036, 0.04187432570345531, 0.00882461000145794, 0.0032401224668318998, 0.00033357632307916605, 0.027878116343490307, 0.0022277299897944304, 0.14333518005540166, 0.1725534334451086, 0.02781629975215046, 0.006909462020702727, 0.005264907420906838, 0.16661437527336345, 0.004325995043009185, 0.003334596880011664, 0.005312727802886718, 0.14024668318996938, 0.0013261408368566844, 0.003504884093891238, 0.006375273363464061, 0.04964922000291588, 0.008290421344219274, 0.09536783787724158, 0.05394372357486515, 0.005505175681586237, 0.005339553870826651, 0.01782067356757545, 0.006710016037323225, 0.05105933809593235, 0.002983525295232541, 0.002940370316372649, 0.0004548768041988629, 0.01208456043154979, 0.000915585362297711, 0.20146260387811635, 0.067196967487972, 0.006158332118384605, 0.025438110511736407, 0.07753783350342616, 0.1273876658405015, 0.009337804344656656, 0.07683452398308792, 0.0070412596588423975, 0.08747164309666132, 0.0038827817466102928, 0.018116926665694706, 0.005017641055547455, 0.004567429654468581, 0.028277008310249308, 0.05271555620352821, 0.004394809739029013, 0.0013343052923166642, 0.00411605190260971, 0.059621519171890944, 0.09073859163143316, 0.01446858142586383, 0.006770666277883074, 0.003425572240851436, 0.0004455459979588861, 0.010401516256013998, 0.005825922146085436, 0.10833882490158916, 0.007584779122321038, 0.016903921854497742, 0.02719580113719201, 0.0304814112844438, 0.02206385770520484, 0.013064295086747339, 0.02696369733197259, 0.009581571657676046, 0.026761918647033093, 0.006510570053943724, 0.021941390873305145, 0.07042659279778393, 0.05437410701268406, 0.1425175681586237, 0.027802303542790494, 0.037690625455605774, 0.0019606356611750987, 0.1095623268698061, 0.06157748942994606, 0.044618749088788455, 0.04955124653739612, 0.03608689313310978, 0.018381688292754043, 0.003404577926811489, 0.015036594255722409, 0.009600233270156, 0.10794693103951014, 0.12447528794284882, 0.0031981338387520046, 0.0074716430966613205, 0.003202799241871993, 0.13437643971424407, 0.006655197550663361, 0.0036693395538708266, 0.049338970695436656, 0.09486863974340283, 0.0015990669193760023, 0.0026604461291733486, 0.051775477474850555, 0.0041347135150896636, 0.005450357194926374, 0.12030325120279925, 0.04581309228750547, 0.0004537104534188657, 0.12425601399620935, 0.025981629975215047, 0.023926519900860182, 0.04423385333138941, 0.0017950138504155123, 0.002661612479953346, 0.0006333284735384167, 0.008449045050298877, 0.000653156436798367, 0.04816678816153958, 0.008625164018078437, 0.0039037760606502403, 0.005228750546726928, 0.004531272780288672, 0.0056672984400058316, 0.00359585945473101, 0.0032179618020119548, 0.0038093016474704767, 0.011452398308791368, 0.002519317684793702, 0.00280390727511299, 0.005572824026826068, 0.004554599795888614, 0.004531272780288672, 0.0035841959469310393, 0.004400641492928998, 0.0036670068523108326, 0.004839189386207902, 0.006258638285464354, 0.004897506925207757, 0.840776789619478, 0.004968654322787578, 0.002886718180492783, 0.0019757982213150604, 0.0018568304417553576, 0.001691208630995772, 0.09009243329931477, 0.14030150167662925, 0.013242746756086894, 0.013746610293045632, 0.027342761335471644, 0.16938912377897652, 0.006607377168683481, 0.01661933226417845, 0.008173786266219566, 0.13297448607668758, 0.0034675608689313307, 0.016641492928998396, 0.011722991689750693, 0.021493512173786266, 0.03430820819361423, 0.10099548039072752, 0.00873596734217816, 0.0018323370753754193, 0.020103222044029742, 0.047197550663362, 0.040833940807697915, 0.03361189677795597, 0.010844729552412887, 0.005544831608106138, 0.0007522962530981193, 0.01525120279924187, 0.00815512465373961, 0.2109648636827526, 0.058258055110074355, 0.007181221752442048, 0.043560868931331105, 0.004058900714389853, 0.10618107595859454, 0.0062399766729844, 0.004835690333867911, 0.02679224376731302, 0.08414637702288964, 0.0030698352529523252, 0.03637498177576906, 0.01592885260242018, 0.017413617145356466, 0.008430383437818923, 0.037231083248286924, 0.03290275550371775, 0.007538125091121154, 0.004500947660008748, 0.05932409972299169, 0.16006764834523984, 0.03309636973319726, 0.007766729844000583, 0.005225251494386936, 0.0006321621227584196, 0.012989648636827526, 0.005274238227146815, 0.1254503571949264, 0.12852719055255868, 0.0035433736696311416, 0.005203090829566993, 0.0019314768916751715, 0.20520775623268697, 0.002509986878553725, 0.00343606939787141, 0.027138649948972155, 0.13926578218399185, 0.004565096952908587, 0.005614812654905963, 0.00874413179763814, 0.004109053797929727, 0.008300918501239247, 0.08270943286193323, 0.002912377897652719, 0.0037066627788307337, 0.06909578655780726, 0.03242805073625893, 0.05237614812654906, 0.04723487388832191, 0.0038991106575302524, 0.006299460562764251, 0.00043388249015891526, 0.020029741944889927, 0.005311561452106721, 0.09334072022160665, 0.022940953491762648, 0.024658988190698353, 0.02901297565242747, 0.03531593526753171, 0.0758023035427905, 0.013711619769645722, 0.021597317393206007, 0.009670214316955824, 0.044728386062108175, 0.010596296836273509, 0.03264382563055839, 0.0604822860475288, 0.05489546581134276, 0.11501851581863246, 0.01837585653885406, 0.026237060796034405, 0.0011255285026971862, 0.08704125965884241, 0.10156349322058608, 0.06660562764251349, 0.023434319871701415, 0.010777081207173057, 0.005409534917626476, 0.003123487388832191, 0.0028762210234728096, 0.0089995626184575, 0.07518297127861205, 0.2314868056568013, 0.002226563639014434, 0.003285610147251786, 0.0027455897361131363, 0.2724537104534189, 0.0016655489138358362, 0.0019209797346551977, 0.0022137337804344656, 0.17690392185449774, 0.0014532730718763668, 0.0024994897215337513, 0.015302522233561744, 0.003441901151771395, 0.015303688584341741, 0.09314593964134713, 0.0017833503426155418, 0.0005108616416387229, 0.017828838023035427, 0.010385187345094037, 0.003168975069252078, 0.01902901297565243, 0.005525003644846187, 0.0010088934246974776, 0.0009272488700976819, 0.036282840064149294, 0.0022977110365942554, 0.0766805656801283, 0.22270418428342326, 0.005283569033386791, 0.007155562035282111, 0.01173582154833066, 0.1715620352821111, 0.003925936725470185, 0.004425134859308937, 0.020040239101909902, 0.14243242455168392, 0.0016737133692958156, 0.0066808572678232975, 0.011980755212130047, 0.012638577052048404, 0.07206065024055984, 0.08115701997375711, 0.00710424260096224, 0.0007278028867181805, 0.02347630849978131, 0.04595538708266512, 0.01481965301064295, 0.013925061962385188, 0.0018125091121154687, 0.00529173348884677, 0.0016340574427759146, 0.03072401224668319, 0.0023746901880740633, 0.25174165330223064, 0.06673392622831317, 0.00878378772415804, 0.03956261845750109, 0.010077270739174807, 0.0844787869951888, 0.00985216503863537, 0.004973319725907567, 0.01893220586091267, 0.11200583175389998, 0.0028715556203528212, 0.004095057588569762, 0.01202391019098994, 0.01756757544831608, 0.014825484764542934, 0.05312961073042717, 0.06746872721971132, 0.003845458521650386, 0.0210806239976673, 0.019443067502551394, 0.08017028721387957, 0.01825572240851436, 0.005365213587986587, 0.01959702580551101, 0.026184575010934536, 0.02474879720075813, 0.002171745152354571, 0.25827321767021433, 0.048050153083539875, 0.01043184137629392, 0.03930485493512174, 0.027640180784370902, 0.03294007872867765, 0.006474413179763814, 0.018314039947514214, 0.015119405161102202, 0.014706516984983233, 0.005494678524566263, 0.03309870243475726, 0.043864120134130345, 0.058996355153812505, 0.06265986295378335, 0.04633328473538417, 0.03790756670068523, 0.0004642076104388394, 0.037849249161685375, 0.08369966467415076, 0.04999679253535501, 0.02392768625164018, 0.010998687855372504, 0.009881323808135296, 0.003867619186470331, 0.012434465665548913, 0.007253535500801866, 0.11106225397288234, 0.17624726636535937, 0.008209943140399476, 0.008390727511299025, 0.012682898381688294, 0.1825653885406036, 0.001538416678816154, 0.004590756670068524, 0.008710307625018223, 0.1299513048549351, 0.002677941390873305, 0.012309666132089225, 0.014087184720804781, 0.01199941682461, 0.031246537396121883, 0.07206648199445984, 0.008254264470039366, 0.0007033095203382417, 0.007034261554162415, 0.006599212713223502, 0.013906400349905234, 0.050098265053214755, 0.007133401370462167, 0.017750692520775622, 0.0008257763522379356, 0.03918821985712203, 0.06015454147834961 ]
                    list
                )
                ()


suiteParseCommon : Test
suiteParseCommon =
    test "Confirm parsing common password list matches reference." <|
        \() ->
            let
                list =
                    parseCommonList commonList
                        |> Result.withDefault (CommonPasswordList Set.empty)
                        |> (\c ->
                                case c of
                                    CommonPasswordList s ->
                                        Set.toList s
                           )
            in
            Expect.equalLists
                [ " ", "!@#$%", "!@#$%^", "!@#$%^&", "!@#$%^&*", "*", ".,m", ".,mn", ".,mnb", "/.,", "/.,m", "/.,mn", "/.,mnb", "/dev/null", "/etc/passwd", "/usr/group", "000", "0000", "00000", "000000", "0000000", "00000000", "0007", "007", "007007", "0213", "0246", "0249", "1", "1022", "1029", "10293", "102938", "106", "10sne1", "111", "1111", "11111", "111111", "1111111", "11111111", "1209", "1212", "121212", "1213", "1214", "1217", "1225", "123", "123098", "123123", "123321", "1234", "12345", "123456", "1234567", "12345678", "1234qwer", "123abc", "123go", "1313", "131313", "1316", "1332", "13579", "1412", "1430", "14430", "1701d", "171717", "1812overture", "1818", "181818", "1900", "1901", "1902", "1903", "1904", "1905", "1906", "1907", "1908", "1909", "1910", "1911", "1912", "1913", "1914", "1915", "1916", "1917", "1918", "1919", "1920", "1921", "1922", "1923", "1924", "1925", "1926", "1927", "1928", "1929", "1930", "1931", "1932", "1933", "1934", "1935", "1936", "1937", "1938", "1939", "1940", "1941", "1942", "1943", "1944", "1945", "1946", "1947", "1948", "1949", "1950", "1951", "1952", "1953", "1954", "1955", "1956", "1957", "1958", "1959", "1960", "1961", "1962", "1963", "1964", "1965", "1966", "1967", "1968", "1969", "1970", "1971", "1972", "1973", "1974", "1975", "1976", "1977", "1978", "1979", "1980", "1981", "1982", "1983", "1984", "1985", "1986", "1987", "1988", "1989", "1990", "1991", "1992", "1993", "1994", "1995", "1996", "1997", "1998", "1999", "1a2b3c", "1chris", "1kitty", "1p2o3i", "1q2w3e", "1qw23e", "1sanjose", "20", "2000", "2001", "2002", "2003", "2004", "2005", "2006", "2007", "2008", "2009", "2010", "2012", "2013", "2014", "2015", "2016", "2017", "2018", "2019", "2020", "2021", "2022", "2023", "2024", "2025", "2026", "2027", "2028", "2029", "2030", "2031", "2032", "2033", "2034", "2035", "2036", "2037", "2038", "2112", "21122112", "22", "2200", "222", "2222", "22222", "222222", "2222222", "22222222", "2252", "2kids", "2welcome", "3", "3010", "3112", "3141", "333", "3333", "33333", "333333", "3333333", "33333333", "3533", "369", "3bears", "4", "4.2bsd", "4.3bsd", "4055", "4077mash", "42bsd", "43bsd", "444", "4444", "44444", "4444444", "44444444", "4788", "4854", "4runner", "5", "5050", "5121", "5252", "54321", "555", "5555", "55555", "555555", "5555555", "55555555", "5683", "57chevy", "5and5", "6262", "6301", "654321", "666", "6666", "66666", "666666", "6666666", "66666666", "6969", "696969", "69zulu4z", "777", "7777", "77777", "777777", "7777777", "77777777", "789456", "7dwarfs", "80486", "8675309", "87654321", "888", "8888", "88888", "888888", "8888888", "88888888", "90210", "911", "911sc", "911scturbo", "911turbo", "92072", "999", "9999", "99999", "999999", "9999999", "99999999", ";lk", ";lk;lk", ";lkasd", "@#$%^&", "a", "a12345", "a1b2c3", "a1b2c3d4", "aa", "aaa", "aaaa", "aaaaa", "aaaaaa", "aaaaaaa", "aaaaaaaa", "aardvark", "aaron", "aarti", "abacab", "abadabdoo", "abbot", "abbott", "abby", "abc", "abc123", "abcd", "abcd123", "abcd1234", "abcde", "abcdef", "abcdefg", "abdenace", "abdol", "abdul", "abdulkaf", "abdullah", "abdur", "abhijit", "abhiram", "abigail", "about", "abracadabra", "abracadaver", "abraham", "abrar", "absolut", "acacia", "academia", "academic", "accept", "access", "accord", "account", "accounts", "ace", "acknak", "acropolis", "action", "active", "actor", "acura", "ada", "adam", "adams", "adel", "adg", "adi", "adib", "adidas", "adine", "adm", "admin", "admin1", "administ", "administrator", "adrian", "adrianna", "adrianne", "adrien", "adrienne", "adrock", "adult", "adventur", "advil", "aeh", "aeneas", "aerobics", "afresh", "africa", "afrid", "after", "again", "agent", "aggie", "aggies", "agnes", "ahidee", "ahmed", "ahmet", "aikman", "aileen", "aimee", "airborne", "aircraft", "airhead", "airplane", "airwolf", "ajai", "ajay", "akhil", "aki123", "akiko", "alain", "alamgir", "alan", "alas", "alaska", "alastair", "alayne", "albany", "albatros", "albatross", "albert", "alberto", "alcapone", "alejandr", "alena", "alert", "alessand", "alex", "alex1", "alexande", "alexander", "alexandr", "alexendr", "alexia", "alexis", "alf", "alfaro", "alfred", "algebra", "ali", "alias", "aliases", "alica", "alice", "alice1", "alicia", "alien", "aliens", "alina", "aline", "alisa", "alison", "all", "allah", "allan", "allegro", "allen", "allison", "allo", "allstate", "aloha", "alok", "alpha", "alpha1", "alphabet", "alpine", "altaf", "altamira", "althea", "altima", "altima1", "alva", "alvin", "always", "alyson", "alyssa", "ama", "amadeus", "amanda", "amanda1", "amar", "amarjit", "amarpree", "amazing", "amber", "amelie", "america", "america7", "ami", "amiga", "amorphous", "amos", "amour", "amril", "amy", "an", "an-jen", "anaconda", "anal", "analog", "anant", "ananth", "anastasi", "anchana", "anchor", "and", "anders", "anderson", "andi", "andre", "andre1", "andrea", "andrea1", "andreas", "andrew", "andrew!", "andrew1", "android", "andromache", "andromed", "andrzej", "andy", "aneliese", "anewpass", "angel", "angel1", "angela", "angela1", "angelika", "angels", "angerine", "angie", "angie1", "angus", "ani", "anil", "animal", "animal house", "animalhouse", "animals", "anis", "anita", "anjana", "anjen", "ann", "anna", "annalena", "annalise", "annamari", "anne", "anneli", "annette", "anni", "annie", "anon", "anonymous", "answer", "antares", "anthony", "anthropogenic", "antoine", "anton", "antonio", "antony", "anu", "anumber1", "anumberone", "anupa", "anupam", "anurag", "anvils", "any", "anything", "apache", "apl", "apollo", "apollo13", "apple", "apple1", "apple2", "appleii", "applepie", "apples", "april", "aptiva", "aqua", "aquarious", "aquarius", "aragorn", "aram", "arash", "arbenz", "archie", "architect", "arctic", "ardent", "are", "arelene", "ari", "aria", "ariadne", "ariane", "ariel", "ariella", "arif", "arijit", "arindam", "aristotle", "arizona", "arjun", "arjunasa", "arlene", "armand", "armando", "armond", "arnold", "aron", "around", "arrow", "arsenal", "arshad", "art", "artemis", "arthur", "artie", "artist", "arty", "arun", "aruna", "arvind", "as", "asad", "asap", "asd", "asdf", "asdf1234", "asdf;lkj", "asdfasdf", "asdfg", "asdfgh", "asdfghj", "asdfghjk", "asdfjkl", "asdfjkl;", "asdlkj", "ashima", "ashish", "ashley", "ashley1", "ashok", "ashraf", "ashton", "ashutosh", "asian", "asjeet", "ask", "asm", "aspen", "ass", "asshole", "assmunch", "asterix", "at", "at&t", "atandt", "ate", "ath", "athanass", "athena", "atlanta", "atmosphere", "atse", "attila", "atul", "audie", "audra", "audrey", "august", "augustin", "aurelius", "austin", "author", "autumn", "avalon", "avatar", "avenger", "avenir", "avi", "avni", "avraham", "away", "away!", "awesome", "ayelet", "aylmer", "azam", "aziz", "azizi", "aztecs", "azure", "babak", "babe", "babes", "babies", "baby", "babydoll", "babylon5", "bacchus", "bach", "backdoor", "backrub", "backup", "badass", "badboy", "badger", "badtimes", "baglady", "bagwoman", "bahram", "bailey", "bakedpotatoe", "baker", "bakshi", "balakris", "balas", "balasubr", "baldo", "balkrish", "ballard", "balls", "bambi", "bamboo", "ban", "banana", "bananas", "banane", "bancroft", "bandit", "bang", "banks", "banzai", "baraka", "barb", "barbara", "barber", "barbie", "barf", "barfer", "barfing", "baritone", "barn", "barnes", "barney", "barney1", "barnie", "barnyard", "baron", "baron harkonnen", "baronharkonnen", "barret", "barrett", "barry", "bart", "bartman", "barton", "basebal", "baseball", "basf", "basful", "basic", "basil", "baskar", "basket", "basketb", "basketba", "basketball", "bass", "bassoon", "bastard", "basuki", "batcave", "batch", "batcomputer", "batman", "batman1", "batmobile", "baystate", "bball", "bbb", "bbbb", "bbbbb", "bbbbbb", "bbbbbbb", "bbbbbbbb", "be", "beach", "beaches", "beagle", "beammeup", "beaner", "beanie", "beans", "bear", "bears", "beast", "beasty", "beater", "beatit", "beatles", "beatrice", "beautifu", "beautiful", "beauty", "beaver", "beavis", "beavis1", "bebe", "because", "becca", "becky", "beef", "been", "beer", "beethoven", "before", "behnam", "belgium", "belize", "bell", "bella", "belle", "bellow", "belmont", "beloved", "ben", "bengt", "benjamin", "benji", "bennet", "bennett", "benny", "benoit", "benson", "bent", "beny", "benz", "beowulf", "beppe", "beresfor", "berhanu", "berkeley", "berlin", "berliner", "berlinwall", "bernard", "bernardo", "bernhard", "bernie", "berry", "bert", "bertha", "beryl", "best", "beta", "betacam", "beth", "bethany", "betsie", "betsy", "better", "bettie", "betty", "beverly", "bfd", "bfi", "bharat", "bhavani", "bhoothap", "biay", "bicameral", "bichnga", "bichon", "bienveni", "big", "bigal", "bigben", "bigbird", "bigboss", "bigboy", "bigbrother", "bigbucks", "bigcock", "bigcocks", "bigdeal", "bigdog", "bigdude", "bigfoot", "biggles", "bigguy", "bighips", "bighouse", "bigjoke", "bigmac", "bigman", "bigmouth", "bigred", "bigroom", "bigsecret", "bigtits", "bigtoe", "biker", "bilbo", "biliamee", "bill", "billc", "billie", "bills", "billy", "billy1", "bim", "bimbo", "bimboe", "bimmer", "bin", "bind", "bing", "bingo", "binky", "binod", "bioboy", "biochem", "biology", "bird", "bird33", "birdie", "birdy", "birget", "birgetta", "birgit", "birthday", "biscuit", "bishop", "bismillah", "bitch", "bitchin", "bitchin'", "bitching", "biteme", "bitter", "biz", "bizhan", "bjorn", "black", "blackboot", "blackie", "blackjack", "bladerunner", "blah", "blaine", "blair", "blake", "blanche", "blaster", "blazer", "bleep", "bleeping", "bleeps", "blinds", "bliss", "blitz", "blizzard", "blonde", "blondie", "blonds", "blonds1", "blood", "bloomcounty", "blow", "blowfish", "blowjob", "blowme", "blowoff", "blss", "blue", "bluebird", "blueblazer", "blueeyes", "bluefish", "bluejean", "blueline", "blues", "bluesky", "bluevelvet", "bmw", "bo", "boat", "bob", "bobbi", "bobbijo", "bobby", "bobcat", "bodyshop", "boeing", "bogart", "bogey", "bogus", "boleslaw", "bombay", "bond", "bond007", "boner", "bong", "bonita", "bonjour", "bonkers", "bonn", "bonnie", "bonsai", "bonzai", "bonzo", "boobie", "boobies", "booboo", "boobys", "booger", "boogie", "book", "bookem", "bookemdanno", "bookit", "boomer", "boon", "booster", "boots", "bootsie", "boozie", "bor", "boris", "bornagai", "bosco", "boss", "boston", "both", "boulder", "bourbon", "bourne", "bourne-again", "bourneagain", "bowling", "boxer", "boxers", "boyd", "boywonder", "bozo", "brad", "brad&janet", "bradford", "bradjanet", "bradley", "bradnjanet", "brain", "braindead", "branch", "brand-n-janet", "brandi", "brandon", "brandy", "branisla", "brasil", "brat", "bravenewworld", "braves", "brazil", "breakout", "breast", "breastfeed", "breasts", "brenda", "brendan", "brenden", "brent", "bret", "breton", "brett", "brewster", "brian", "brickles", "brickout", "bridge", "bridges", "bridget", "bridgett", "briefcase", "bright", "bring", "brinkley", "britain", "broadway", "brokenheart", "broker", "bromberg", "bronco", "broncos", "bronte", "brooke", "brooks", "brothel", "brother", "brothers", "brown", "browns", "bruce", "bruno", "brutus", "bryan", "bryant", "bryce", "bryn", "bsd", "bsd4", "bsdunix", "bubba", "bubba1", "bubbah", "bubbahlah", "bubbalah", "bubble", "bubbles", "buck", "buckaroo", "bucks", "budd", "buddah", "buddha", "buddhist", "buddy", "budgie", "budlite", "buffalo", "buffett", "buffy", "bugs", "bugsbunny", "bugsy", "bull", "bulldog", "bullet", "bulls", "bullshit", "bumbling", "bung", "bunny", "bunnyrabbit", "burgess", "burke", "burns", "burton", "busalacc", "business", "buster", "but", "butch", "butler", "butt", "butter", "butterfly", "buttfuck", "buttfucker", "butthead", "button", "buttons", "buy", "buzz", "by", "byoung", "byoungin", "byron", "byteme", "byung", "c++", "c00per", "c3po", "cabernet", "cabinboy", "cactus", "cad", "cadat", "cadcam", "cadweld", "caesar", "caip", "caipcad", "caitlin", "cal", "caleb", "calendar", "calgary", "caliban", "californ", "california", "call", "calvin", "calvin1", "camaro", "camay", "camel", "camera", "cameron", "camilla", "camille", "camlin", "campanile", "campbell", "camping", "can", "canada", "canced", "cancer", "candace", "candi", "candy", "canela", "cannon", "cannonda", "cannondale", "canon", "cantor", "canucks", "canute", "capfast", "captain", "captian", "car", "carbon", "card", "cardinal", "carebear", "caren", "carey", "carl", "carla", "carlena", "carlo", "carlos", "carlyle", "carlyn", "carmen", "carnage", "carol", "carol1", "carole", "caroleen", "carolie", "carolina", "caroline", "carolyn", "carrie", "carrol", "carrot", "carry", "carson", "carter", "cartman", "carver", "cary", "caryl", "caryn", "cascade", "cascades", "casey", "cash", "cashbox", "casio", "caspar", "casper", "cassie", "castle", "cat", "catalina", "catalog", "catch22", "catfish", "catherin", "catherine", "cathi", "cathleen", "cathy", "catnip", "cats", "catwoman", "cayuga", "ccc", "cccc", "ccccc", "cccccc", "ccccccc", "cccccccc", "cecil", "cecile", "cecilia", "cecily", "cedic", "celeste", "celia", "celica", "celine", "celtics", "cement", "center", "cerebus", "cerulean", "cesar", "cesare", "cessna", "cfi", "cfj", "cgj", "chad", "chai", "chain", "chainsaw", "chakkala", "challeng", "challenge", "chameleon", "champion", "champs", "chan", "chance", "chand", "chandler", "chandra", "chandram", "chandras", "chanel", "chanequa", "chang", "change", "changed", "changeit", "changeme", "changethis", "changho", "changkyu", "channel", "channi", "chanshin", "chantal", "chao", "chao-yan", "chaofeng", "chaos", "chapman", "chardonnay", "charger", "charity", "charles", "charlie", "charlie1", "charlott", "charlotte", "charming", "charon", "chas", "chat", "chau", "checkin", "checkov", "chedsada", "cheese", "cheesecake", "cheif", "chelsea", "chelsea1", "chem", "chemistry", "chen", "cheng", "cheow", "cheow-to", "cherala", "cherry", "cheryl", "chess", "chester", "chester1", "cheung", "chevy", "chevy1", "chi", "chi-pang", "chi-shun", "chi-tai", "chi-wang", "chi-yao", "chia", "chia-hua", "chia-lin", "chia-yin", "chia-yu", "chiara", "chicago", "chicken", "chico", "chiefs", "chien", "chihsing", "childsplay", "chilin", "chin", "chin-w", "china", "chinacat", "ching", "ching-en", "ching-li", "ching-me", "chinook", "chinpan", "chip", "chipper", "chiquita", "chisheng", "chldrn", "chloe", "cho", "chocolat", "chocolate", "chol", "choong", "choong-h", "chop", "chopsticks", "chou", "chouette", "chris", "chris1", "chris123", "chrispen", "chriss", "chrissy", "christ", "christ1", "christia", "christian", "christie", "christin", "christina", "christine", "christmas", "christop", "christoph", "christopher", "christy", "chronos", "chu", "chuck", "chucky", "chuen-ch", "chuen-ts", "chun", "chun-lin", "chun-she", "chun-yu", "chung", "chung-na", "chung-pi", "chung-ya", "chungen", "chungyen", "chuong", "church", "churn-hu", "cic", "cicero", "cigar", "cimarron", "cindelyn", "cinder", "cindi", "cindy", "cindy1", "cinema", "circuit", "cirque", "cirrus", "civic", "civil", "claire", "clambake", "clancy", "clapton", "clarence", "clarisa", "clarissa", "clark", "clarkson", "class", "classic", "classroo", "classroom", "claude", "claudel", "claudia", "cleaner", "cleanfight", "cleanroom", "cleat", "cleavage", "cleo", "cliff", "clifford", "clifton", "clint", "clinton", "clipper", "clit", "clitoris", "clock", "cloclo", "closefriend", "cloud", "clover", "clueless", "cluster", "clusters", "coatamundi", "coatimundi", "cobain", "cobra", "cocacola", "cocakola", "cock", "coco", "code", "codename", "cody", "coe", "coffee", "cohen", "cokacola", "coke", "cokeisit", "colby", "cold", "coldcuts", "coldshoulder", "coldwar", "coleman", "colette", "colin", "colleen", "college", "collette", "collins", "color", "colorado", "colors", "colour", "colt45", "coltrane", "columbia", "comandur", "combination", "come", "comedienne", "comeon", "comet", "commander", "commrades", "communication", "compaq", "compton", "compuserve", "compute", "computer", "comrade", "comrades", "concept", "concorde", "condo", "condom", "condoms", "confiden", "confused", "connect", "conner", "connie", "conrad", "console", "conspiritu", "consulta", "consulti", "content", "control", "convex", "cook", "cookie", "cookies", "cooking", "cool", "coolbean", "coolman", "cooper", "cooter", "copper", "cora", "coralyn", "cordelia", "corey", "corinna", "corinne", "corky", "corlene", "cornelia", "cornelius", "cornflake", "corona", "corrado", "corvette", "corwin", "cosmic", "cosmo", "cosmos", "cougar", "cougars", "could", "country", "couples", "courier", "courtney", "couscous", "coventry", "cowboy", "cowboys", "cows", "cox", "coyote", "crack1", "cracker", "craig", "crapp", "craps", "crashcourse", "crawford", "create", "creation", "creative", "credit", "creosote", "crescent", "cretin", "crew", "cricket", "criminal", "cristina", "cronus", "cross", "crow", "crowley", "crug", "cruise", "crusader", "crystal", "cs-ee", "csc298", "csc412", "cscie-ci", "csee", "cshrc", "cthreepo", "cthulhu", "cuda", "cuddles", "cuervo", "cum", "cunningham", "cunt", "cuong", "cupcake", "curmudgeon", "current", "curt", "curtis", "customer", "custsup", "cut", "cutdown", "cutie", "cutiepie", "cutlass", "cyber", "cyberpunk", "cyclone", "cynthia", "cyrano", "cyril", "daddy", "daebum", "daedalus", "daehyun", "daemon", "daemonic", "daemons", "dagger", "dagger1", "daily", "dain", "daisie", "daisy", "dakota", "dale", "dalibor", "dalit", "dallas", "daltry", "dalu", "dame", "damien", "dammit", "damogran", "damon", "damrongs", "dan", "dana", "dance", "dancer", "dane", "dangermouse", "danh", "daniel", "daniel1", "danielle", "danna", "danni", "danny", "dante", "daphne", "dapper", "daqing", "darin", "dark1", "darkman", "darkstar", "darrell", "darren", "darrin", "darrow", "darryl", "darth", "darthvader", "darwei", "darwin", "daryl", "daryouch", "dasha", "dat", "data", "data1", "database", "datatrain", "datoo", "dave", "david", "david1", "davidovi", "davids", "davis", "dawit", "dawn", "day", "daytek", "ddanielrod", "dday", "ddd", "dddd", "ddddd", "dddddd", "ddddddd", "dddddddd", "de'an", "dead", "dead-head", "deadahead", "deadhead", "deadhed", "dean", "deanna", "death", "deathstar", "deb", "debasish", "debbie", "deborah", "debra", "december", "decker", "dedhead", "dedhed", "dedi", "dee", "deedee", "deepak", "deepfreeze", "deepsea", "deepsix", "deepspace", "deepthroat", "deeznuts", "def", "default", "defense", "defoe", "dekai", "delano", "delete", "deliver", "delnaz", "delois", "delta", "deluge", "demeter", "demo", "demon", "demonic", "demons", "denali", "denis", "denise", "dennis", "denny", "denver", "depeche", "dept", "dequin", "derek", "derluen", "derrek", "desaree", "desert", "design", "desiree", "deskjet", "desktop", "desmond", "desperate", "destiny", "detleff", "detroit", "deutsch", "dev", "devadmin", "develop", "deven", "device", "devil", "devilinside", "devine", "devon", "dew", "dewayne", "dewey", "dewydecimal", "dexter", "dgj", "dhan", "dharma", "dharmara", "dhiraj", "diablo", "diag", "diags", "dial", "dial-in", "dial-up", "dialin", "dialup", "diamond", "diamonds", "dian", "diana", "diane", "diann", "dianne", "diaz", "dick", "dickens", "dickhead", "dicktracy", "did", "diego", "diesel", "diet", "dieter", "digger", "digital", "digital1", "dilbert", "dilip", "dillweed", "dim", "dimitris", "dimwit", "dina", "dinesh", "dinner", "dipak", "diplomac", "diplomacy", "dipper", "dipstick", "dipta", "direct1", "director", "dirk", "disc", "discbrakes", "discjockey", "disco", "discovery", "disk", "disney", "dixie", "dixon", "do", "doan", "doc", "doctor", "document", "dodger", "dodgers", "does", "dog", "dogbert", "dogcatcher", "dogfight", "doggie", "doggy", "doit", "doitnow", "dollar", "dollars", "dolly", "dolphin", "dolphins", "domain", "domenico", "dominic", "dominick", "dominique", "domino", "don", "don't", "donald", "done", "dong", "dongming", "donkey", "donn", "donna", "dontknow", "doobie", "doofus", "doogie", "dookie", "doom", "doom2", "doon", "doors", "dopey", "dorab", "dorai", "dorcas", "dori", "doris", "dorit", "dork", "dorothy", "dos", "double", "doudou", "doug", "dougie", "douglas", "down", "downtown", "dr.demento", "draft", "dragon", "dragon1", "dragonfl", "dragonfly", "dragons", "draw", "draxo", "drazen", "dream", "dreamer", "dreams", "dreamscape", "drew", "drillpress", "drink", "dripping", "driver", "drizzt", "drno", "drop", "drop dead", "dropdead", "drought", "drugnig", "drum", "drums", "dryden", "duane", "duck", "duckie", "ducksbreath", "ducksoup", "dude", "dudley", "duke", "duke leto", "dukeleto", "dulce", "dumbass", "dunbar", "duncan", "dundee", "dune", "dungeons", "dunn", "duplicate", "dustin", "dusty", "dutch", "dutchess", "dvlinside", "dwain", "dwane", "dwayne", "dwight", "dylan", "e", "e-mail", "each", "eager", "eagle", "eagle1", "eagles", "earl", "earth", "easier", "easter", "eastern", "easton", "easy", "easycome", "easygo", "easylay", "eat", "eatme", "eatshit", "eatshitand", "eckart", "eclipse", "ed", "eddie", "eden", "edgar", "edges", "edinburgh", "edith", "edmund", "edouard", "eduard", "eduardo", "edward", "edwards", "edwin", "edwina", "ee-cs", "eecs", "eee", "eeee", "eeeee", "eeeeee", "eeeeeee", "eeeeeeee", "eeyore", "effie", "egghead", "eiderdown", "eieio", "eight", "eileen", "einstein", "eirik", "eka", "ekaterin", "eladio", "elaine", "elanor", "electric", "element", "elena", "eleni", "elephant", "eli", "elias", "elina1", "eliot", "elisabet", "elissa", "elizabet", "elizabeth", "ella", "ellen", "ellie", "elliot", "elliott", "ellis", "elmira", "elmo", "elmootaz", "elmstree", "eloise", "elsie", "elvin", "elvira", "elvis", "elwood", "elwyn", "email", "emerald", "emil", "emile", "emilio", "emily", "emmanuel", "emmi", "emmitt", "emory", "empire", "emptyhanded", "emptyheaded", "enda", "endah", "enemy", "energy", "engage", "engine", "engineer", "england", "enigma", "enrique", "enter", "enterprise", "entropy", "enzo", "enzyme", "eratea", "erenity", "erh", "erhard", "eric", "eric1", "erica", "erich", "erik", "erika", "erin", "erling", "ernest", "ernesto", "ernie", "ernie1", "ernst", "erotic", "ersatz", "ertyuiop", "ervan", "escort", "escort1", "esfandia", "esmond", "espanol", "establish", "estate", "estelle", "esther", "et", "etaoin", "etaoin shrdlu", "etaoinshrdlu", "etee", "eternity", "ethan", "etoile", "eucc", "euclid", "eugene", "eung", "eunji", "europe", "eva", "evan", "evans", "eve", "evelyn", "everafter", "every", "evie", "exavier", "excaliber", "excalibu", "excalibur", "excel", "except", "explore", "explorer", "exponent", "export", "express", "extension", "exxxtreme", "eyal", "face", "faculty", "fail", "fairview", "fairway", "fairyring", "faith", "falcon", "falsestart", "family", "family1", "fancy", "fang", "far", "farah", "faretheewell", "farflung", "fargone", "farhad", "farmer", "farming", "farout", "farrell", "farside", "faruk", "fasihudd", "fast", "fastbreak", "fastlane", "fataneh", "fatboy", "fatcat", "faust", "faye", "fayez", "fearless", "february", "feedback", "feedme", "felicia", "feliks", "felipe", "felix", "fender", "fenris", "fereydoo", "fergus", "ferguson", "fermat", "ferrari", "ferret", "ferris", "fff", "ffff", "fffff", "ffffff", "fffffff", "ffffffff", "fgh", "fiction", "fidel", "fidelity", "field", "figaro", "figleaf", "file", "filesyst", "finance", "find", "finite", "finn", "fiona", "fire", "fireball", "firebird", "fireman", "firenze", "firewalk", "first", "fish", "fish1", "fisher", "fishers", "fishes", "fishhead", "fishie", "fishing", "fist", "fitness", "five", "flakes", "flamingo", "flanders", "flash", "fld", "flea", "fleming", "fletch", "fletcher", "fleurs", "flight", "flip", "flipper", "float", "flophouse", "flores", "florida", "florida1", "flow", "flower", "flowerpot", "flowers", "floyd", "fluffy", "flute", "fly", "flyaway", "flyboy", "flyer", "flyers", "flying", "flyingfuck", "flyingleap", "foghorn", "fon", "fong", "foo", "foobar", "foobaz", "fool", "foolproof", "foot", "footbal", "football", "for", "foram", "forbidden", "ford", "forearm", "foresight", "forest", "forkedtounge", "format", "forrest", "forsythe", "fortune", "forward", "foster", "foulplay", "found", "fountain", "four", "fourier", "fourwheel", "fouryears", "fox", "foxtrot", "foxylady", "fozzie", "framemaker", "france", "frances", "francesc", "francesco", "francine", "francis", "francisc", "francois", "frank", "franka", "frankenfurter", "frankie", "franklin", "franknfurter", "frat", "freak1", "freakbrothers", "fred", "freddie", "freddy", "frederic", "frederik", "fredric", "free", "freebird", "freedom", "freeman", "french", "french1", "frenchfries", "freshbread", "freshmeat", "friday", "friedric", "friend", "friends", "frighten", "frisco", "fritz", "frodo", "frog", "frog1", "froggie", "froggies", "froggy", "frogs", "from", "front242", "frontier", "froshmeat", "frosty", "ftp", "fubar", "fuck", "fuck-off", "fuck-you", "fuckaduck", "fucked", "fuckem", "fucker", "fuckface", "fucking", "fuckit", "fuckleg", "fuckme", "fuckoff", "fucku", "fuckyou", "fugazi", "full", "fun", "function", "fungible", "funguy", "funky", "funny", "funtime", "furball", "future", "fuzbat", "fuzz", "fuzzball", "fuzzbat", "gabby", "gabriel", "gabriell", "gaby", "gad", "gadi", "gaelic", "gage", "gail", "gala", "galaga", "galaxian", "galaxy", "galen", "galileo", "gamal", "gambit", "gambler", "games", "gammaphi", "ganapath", "gandalf", "ganja", "gaoyuan", "garcia", "garden", "gardner", "garfield", "garfunkel", "gargoyle", "garlic", "garnet", "garp", "garr", "garrett", "garry", "garth", "gary", "gash", "gasman", "gaston", "gateway", "gateway2", "gator", "gator1", "gatt", "gauss", "gautam", "gave", "gaven", "gavriel", "gedanken", "gemini", "gene", "general", "genesis", "genius", "geof", "geoff", "geoffrey", "georg", "george", "george1", "georgia", "georgina", "gerald", "gerard", "gerardo", "gerd", "gergory", "german", "germany", "germany1", "geronimo", "gerry", "gert", "gertrude", "get", "get fucked", "get laid", "get lost", "get stuffed", "getfucked", "getlaid", "getlost", "getlost!", "getlucky", "getout", "getstuffed", "ggeorge", "ggg", "gggg", "ggggg", "gggggg", "ggggggg", "gggggggg", "ghandi", "gholamal", "ghost", "giancarl", "giants", "gibbons", "gibson", "giffy", "gigi", "gil", "gilbert", "gilgamesh", "gilles", "gilman", "gina", "ginger", "gingers", "gino", "giovanne", "giridhar", "girl", "girls", "giselle", "giuseppe", "give", "gizmo", "glacier", "gladys", "glen", "glenda", "gleneagles", "glenn", "glider1", "global", "gloria", "gmoney", "gnu", "gnuemacs", "gnus", "go", "go away", "go away!", "go fuck yourself", "go jump in a lake", "go to hell", "goahead", "goalie", "goat", "goaway", "goaway!", "goblin", "goblue", "gocougs", "god", "godflesh", "godiva", "godzilla", "goes", "goethe", "gofish", "goforit", "gofuckyourself", "gohome", "going", "gojumpinalake", "gold", "golden", "goldfinger", "goldfish", "goldie", "golf", "golfer", "golfing", "gollum", "gone", "gonorrhea", "gonzales", "gonzalez", "gonzo", "goober", "good", "good-luck", "goodafternoon", "goodevening", "goodfight", "goodgrief", "goodjob", "goodluck", "goodmorning", "goodtimes", "goodwife", "goofus", "goofy", "goonight", "goose", "gopalon", "gopher", "gopinath", "gordan", "gordon", "gorgeous", "gorges", "gorog", "gosling", "goson", "gostraight", "got", "goto", "goto hell", "gotohell", "gouge", "gould", "gowest", "grace", "graeme", "graham", "grahm", "grail", "gramps", "grandma", "grant", "graphic", "graphics", "grateful", "gratefuldead", "gravis", "gray", "graymail", "greasyspoon", "great", "greed", "green", "greenday", "greenline", "greeting", "greg", "greg1", "gregg", "gregory", "gremlin", "gremlins", "greta", "gretal", "gretchen", "grete", "gretel", "gretzky", "griffey", "griffin", "gripe", "grissom", "grizzly", "groovy", "group", "grover", "grow", "grumpy", "gryphon", "gsite", "gucci", "guenter", "guess", "guest", "guido", "guillerm", "guinness", "guitar", "guitar1", "gulukota", "gumby", "gumption", "gunner", "guntis", "guozhong", "gupi", "gurjot", "guru", "gus", "gustavo", "guy", "gwen", "gymnast", "h2opolo", "ha", "hack", "hacked", "hacker", "had", "hafidh", "haftan", "haggis", "haha", "hai", "haibo", "hailey", "hairbag", "hairball", "hairil", "hakan", "hal", "hal9000", "hall", "hallelujah", "hallo", "halloween", "hallowell", "halt", "hamid", "hamilton", "hamlet", "hamlin", "hammer", "hammered", "hammond", "hampton", "hamster", "han-gyoo", "handily", "handwave", "handwaving", "hank", "hanna", "hannah", "hans", "hansel", "hansolo", "hanson", "hanspete", "hao", "happening", "happy", "happy1", "happy123", "happyday", "happyending", "hard", "hard2see", "hardcore", "harddisk", "hardi", "hardison", "hardware", "harkara", "harkonnen", "harlan", "harley", "harley1", "harlots", "harmony", "haro", "harold", "harriet", "harris", "harrison", "harrold", "harry", "harue", "haruo", "harvard", "harvey", "has", "hasok", "hassan", "hatton", "hauhua", "have", "havivah", "hawaii", "hawk", "hawkeye", "hawkeye1", "hayes", "hazel", "he", "he'sdead", "he'sdeadjim", "health", "health1", "hean", "heart", "heartbreak", "hearts", "heat", "heather", "heather1", "heather2", "heaven", "hebrides", "hector", "hedgehog", "hee", "heeralal", "heesung", "heidi", "heike", "heikki", "heinlein", "heinrich", "heinz", "helen", "helena", "helene", "helge", "hell", "hello", "hello1", "hello123", "hello8", "hellohello", "help", "help123", "helper", "helpme", "hemant", "henderson", "hendrix", "henning", "henry", "her", "herb", "herbert", "here", "herman", "hermes", "hernandez", "herpes", "hershey", "herve", "herzog", "hesdead", "hesdeadjim", "heung", "hewlett", "heydude", "heythere", "hhh", "hhhh", "hhhhh", "hhhhhh", "hhhhhhh", "hhhhhhhh", "hiawatha", "hibernia", "hidden", "highland", "highlife", "hilarie", "hilbert", "hilda", "hill", "hillary", "hillel", "him", "hiroguch", "hiroki", "hiroo", "hiroshi", "hiroyuki", "his", "histoire", "history", "hitchcock", "hithere", "hitler", "hoa", "hoang", "hobbes", "hobbit", "hockey", "hockey1", "hocus", "hocus pocus", "hocus-pocus", "hocuspocus", "hok", "hola", "hold", "hole", "holiday", "holly", "holy", "holy grail", "holygrail", "holyshit", "homayoum", "home", "homebrew", "homer", "homerj", "homework", "hon", "honda", "honda1", "honey", "hong", "hongkong", "hongphuc", "hongtao", "honkey", "hoodlum", "hooker", "hookers", "hoops", "hoosier", "hooters", "hooters2011", "hootie", "hope", "hopscotch", "horizon", "hornet", "hornets", "horny", "horror", "horse", "horses", "horus", "hosanna", "hosannah", "hosehead", "host", "hot", "hotdog", "hotlips", "hotrod", "hottip", "houcine", "house", "housewife", "houston", "how", "howard", "howard93", "howell", "howie", "hplab", "hsin", "hsiuwen", "hspice", "huang", "huasheng", "hubba", "hubbahubba", "hubert", "hudson", "huey", "hugh", "hughes", "hugo", "hugues", "huiying", "hummer", "hundt", "hung", "hungmok", "hunter", "hunting", "huong", "hurt", "husband", "huskers", "huskies", "hustler", "hutchins", "huu", "huyen", "huzur", "hwansoo", "hydrogen", "hydroxyl", "hye", "hyman", "hyo", "hyon", "hyoung", "hyuk", "i", "i'mok", "i'mokay", "iabg", "ian", "ib6ub9", "ibanez", "ibeleive", "ibelieve", "ibm", "ibmpc", "ibmpcat", "ibmpcxt", "ibmsux", "ibrahim", "icap", "icecream", "iceman", "icon", "identical", "idiot", "idontknow", "if", "if6was9", "iforget", "iforgot", "ignacio", "ignatius", "iguana", "ihacked", "ihao", "ihateyou", "ihtfp", "iii", "iiii", "iiiii", "iiiiii", "iiiiiii", "iiiiiiii", "ikonas", "ikuo", "ilan", "ilmari", "iloveu", "iloveyou", "ilya", "image", "imagine", "imbroglio", "imin", "imok", "imokay", "impact", "impala", "imperial", "imsl", "in", "include", "inderpal", "indian", "indiana", "indigo", "indira", "indonesia", "indra", "info", "informix", "ingemar", "ingmar", "ingo", "ingres", "ingress", "ingrid", "ingvar", "inheritthewind", "inigo", "inna", "innocent", "innocuous", "insane", "insertion", "inside", "insight", "inst", "install", "instruct", "integra", "integral", "intel", "intercourse", "interleaf", "intern", "internet", "interracial", "into", "intrepid", "invinoveritas", "invisible", "invite", "inxs", "ioana", "iomega", "iong", "ira", "ireland", "irene", "irenee", "irfan", "irina", "iris", "irish", "irishman", "irlande", "irma", "irmeli", "ironman", "irulian", "irving", "is", "isa", "isaac", "isabel", "isabelle", "isajoke", "isel", "ishmael", "isi", "isidore", "isil", "isis", "island", "ismail", "israel", "isreal", "issam", "isto", "it", "it'sajoke", "it'sok", "it'sokay", "italia", "italy", "its", "itsajoke", "itsok", "itsokay", "itsy-bitsy", "itsybitsy", "itty-bitty", "ittybitty", "iuytrewq", "ivan", "ivy", "izzy", "j0ker", "j1l2t3", "jack", "jackie", "jackie1", "jackolantern", "jackson", "jacob", "jacqueline", "jacques", "jade", "jae", "jaeger", "jaejin", "jagger", "jaguar", "jahanshi", "jai", "jaik", "jaikne", "jaikumar", "jaime", "jain", "jake", "jakey", "jakov", "jamaica", "james", "james1", "jamesbond", "jamie", "jamilah", "jamison", "jamjam", "jan", "jana", "janaki", "jane", "janek", "janel", "janet", "janice", "janie", "jann", "janna", "janny", "january", "janvier", "jany", "japan", "japanese", "japon", "jared", "jasho", "jashvant", "jasmin", "jasmine", "jason", "jason1", "jaspal", "jasper", "jatin", "javed", "javier", "jaws", "jay", "jayanta", "jayanth", "jayne", "jayson", "jazz", "jean", "jean-baptiste", "jean-claude", "jean-francois", "jean-michel", "jean-pierre", "jean-yves", "jeananda", "jeanclaude", "jeanette", "jeanfrancois", "jeanine", "jeanmichel", "jeanne", "jeannie", "jeanpierre", "jeanyves", "jed", "jedi", "jeepcj", "jeepcj7", "jeepster", "jef", "jeff", "jeffery", "jeffrey", "jeffrey1", "jehan", "jello", "jellystone", "jen", "jenifer", "jenkins", "jenn", "jenni", "jennie", "jennifer", "jenny", "jenny1", "jens", "jensen", "jer", "jerald", "jeremy", "jerome", "jerric", "jerrimy", "jerry", "jersey", "jesse", "jesse1", "jessica", "jessie", "jester", "jesus", "jesus1", "jesuschrist", "jethro", "jethroh", "jethrotull", "jetta1", "jeudi", "jewels", "ji", "jiachen", "jian", "jianli", "jiann", "jianping", "jianwen", "jie", "jihong", "jikun", "jill", "jim", "jimbo", "jimbob", "jimi", "jimmin", "jimmy", "jin", "jing", "jinsheng", "jiong", "jiseong", "jitendra", "jixian", "jjj", "jjjj", "jjjjj", "jjjjjj", "jjjjjjj", "jjjjjjjj", "jkl123", "jkl;", "jkm", "jnye", "joan", "joanie", "joann", "joanna", "joanne", "joaquim", "jocelyn", "jody", "joe", "joel", "joelle", "joena", "joerg", "joey", "johan", "johann", "johanna", "johanna1", "john", "john316", "johnlennon", "johnny", "johnson", "join", "join for free", "joinforfree", "joji", "jojo", "joker", "joker1", "jole", "jon", "jonathan", "jonell", "jones", "jong", "jong-i", "jonggu", "joni", "jonny", "jordan", "jordan23", "jordie", "jorean", "jorge", "jose", "josee", "joseph", "josh", "joshua", "josiah", "josie", "jour", "journey", "joy", "joyce", "jsbach", "juan", "jubilee", "judas", "judi", "judianto", "judicael", "judith", "judy", "juggle", "juhani", "jui", "jui-fen", "juicy", "juillet", "juin", "julayne", "jules", "juli", "julia", "julia2", "julian", "juliana", "juliann", "julie", "julie1", "julien", "julienne", "juliet", "juliette", "july", "jumanji", "jumbo", "jumeaux", "jump", "jump in a lake", "jumpinalake", "jun", "june", "junebug", "jungle", "juni", "junior", "juniper", "juping", "jupiter", "jussi", "just", "justdoit", "juste", "justforthe", "justice", "justice4", "justin", "justin1", "justine", "jutta", "jvnc", "jyh", "kacy", "kadosh", "kai", "kaka", "kakogawa", "kal007", "kalamazo", "kalappa", "kali", "kalyan", "kalyn", "kam", "kan", "kang", "kangaroo", "kaos", "kara", "karalee", "karen", "karen1", "karie", "karin", "karina", "karine", "karl", "karma", "kary", "karyn", "kasey", "kashtan", "kat", "kate", "katerina", "katherin", "katherine", "kathi", "kathleen", "kathreen", "kathrine", "kathryn", "kathy", "kati", "katie", "katie1", "katina", "katrina", "katsufum", "katydid", "kaveh", "kay", "kayla", "kaylen", "kcin", "kecia", "kee", "keep", "keeper", "keepout", "kees", "keh", "keith", "keith1", "keller", "kelley", "kelly", "kelly1", "kelsey", "ken", "kendall", "kenji", "kennedy", "kenneth", "kenny", "kenobi", "kent", "kenton", "kenzo", "kerala", "kerberos", "keri", "kermit", "kernel", "kerri", "kerrie", "kerry", "kerrya", "keshav", "kester", "ketan", "ketchup", "kevin", "kevin1", "kewl", "key", "keyboard", "keypad", "khan", "khanh", "khayroll", "khoanh", "khoi", "khong", "khosrow", "khueh", "khueh-ho", "khurshee", "kian", "kiang", "kianusch", "kiat", "kidder", "kids", "kieu", "killer", "killme", "kilroy", "kim", "kimberly", "kimmo", "kimon", "kind", "kinder", "king", "kingandi", "kingdom", "kingfish", "kinglear", "kings", "kinson", "kip", "kiran", "kirk", "kirkland", "kirsten", "kiss", "kissa2", "kissme", "kitkat", "kitten", "kitten12", "kittens", "kitty", "kittycat", "kiwi", "kjhgfdsa", "kkk", "kkkk", "kkkkk", "kkkkkk", "kkkkkkk", "kkkkkkkk", "klaus", "kleenex", "klingon", "klingons", "knickers", "knicks", "knight", "knights", "know", "knute", "koala", "koichi", "koji", "kokakola", "koko", "kombat", "kongjoo", "konrad", "kook", "korda", "kraig", "kramer", "kris", "krishna", "krishnam", "krista", "kristen", "kristi", "kristie", "kristin", "kristina", "kristine", "kristy", "krystal", "krystyna", "kun", "kuo", "kurt", "kwan", "kwang", "kwok", "kwong", "kyahn", "kyeongso", "kyle", "kyra", "lab1", "labtec", "lacrosse", "laddie", "ladies", "ladle", "lady", "ladybug", "lager", "laid", "lake", "lakers", "lakota", "lakshman", "lalit", "lalith", "lalitha", "lam", "lambda", "lambert", "lamer", "lamination", "lan", "lana", "lance", "lancer", "landry", "lapin", "lara", "larissa", "larkin", "larry", "larry1", "larson", "laser", "laserjet", "lassie1", "last", "lastango", "lasttango", "latenight", "laugh", "laura", "laurae", "lauramae", "laurel", "lauren", "laurence", "laurent", "laurenz", "laurie", "laurinda", "laury", "law", "lawrence", "lawson", "lawyer", "lazare", "lazarus", "lea", "leader", "leaf", "leah", "leann", "leanne", "lebesgue", "leblanc", "led-zeppelin", "leddzeppelin", "ledzep", "ledzepp", "ledzeppelin", "lee", "legal", "legend", "leger", "lehi3b15", "leison", "leisure", "leland", "lemon", "len", "lena", "lennon", "lenore", "leo", "leon", "leonard", "leonce", "leonid", "leroy", "les", "lesbian", "leslie", "lespaul", "lestat", "lester", "let", "leticia", "letlive", "letmein", "leto", "letsgo", "letter", "letters", "lev", "lewis", "lexus1", "lez", "li", "lib", "liberty", "libra", "library", "lick", "licker", "licorne", "lien", "liew", "life", "light", "lights", "like", "lillian", "lilly", "lily", "lima", "limited", "lin", "linc", "lincoln", "linda", "lindsay", "lindsey", "lindy", "ling", "linh", "lion", "lionel", "lionking", "lions", "lisa", "lise", "lisp", "lissabon", "list", "litterbox", "little", "littlehouse", "littleshit", "littleshop", "littletoe", "live", "liveandletlive", "livenletlive", "liverpoo", "liverpool", "lives", "livia", "living", "liwana", "liz", "liza", "lizard", "lizzy", "ljf", "ljiljana", "lkj", "lkjasd", "lkjh", "lkjhg", "lkjhgf", "lkjhgfds", "lkjlkj", "llewelly", "lll", "llll", "lllll", "llllll", "lllllll", "llllllll", "lloyd", "lmnop", "loch", "lock", "lockout", "log", "logan", "logger", "logical", "login", "logos", "logout", "lois", "loislane", "loke", "loki", "lola", "lolita", "lolopc", "london", "lonely", "lonestar", "long", "longcock", "longer", "longhair", "longhorn", "longrest", "longtounge", "look", "looney", "loose", "loosing", "lopez", "loren", "lorenzo", "loretta", "lori", "lorie", "lorin", "lorna", "lorraine", "lorrie", "lory", "loser", "lost", "lotfi", "lotus", "lotus123", "lou", "louie", "louis", "louisa", "louise", "lounette", "lourdes", "love", "lovely", "loveme", "lover", "loverboy", "lovers", "loveyou", "lowgrade", "lowlife", "lp", "lpadmin", "lsd", "ltte", "luana", "luc", "lucas", "lucia", "lucie", "lucien", "lucifer", "lucille", "lucky", "lucky1", "lucky14", "luckybreak", "luckylady", "lucy", "luigi", "luis", "luiz", "luke", "lulu", "lumiere", "lunarlander", "lundi", "lune", "lung", "luong", "luther", "lydia", "lydie", "lyle", "lyndon", "lynette", "lynn", "lynne", "m", "m1911a1", "maarten", "mac", "macha", "machine", "macintos", "macintosh", "mack", "macross", "macse30", "macsyma", "mad", "madboy", "maddie", "maddock", "maddog", "made", "madeleine", "madelene", "madeline", "madhu", "madhusud", "madison", "madman", "madmax", "madoka", "madonna", "mady", "magdalen", "maggie", "maggot", "magic", "magic1", "magique", "magnum", "mahbuba", "mahesh", "mahlon", "mahmoud", "mai", "maia", "maiden", "mail", "mailer", "mailinglist", "mailman", "maine", "mainsail", "maint", "major", "majordom", "majordomo", "make", "makebread", "makedrugs", "makeit", "makeitso", "makelove", "makeme", "makemyday", "makepeace", "makewar", "makingit", "makinglove", "makoto", "malcolm", "malcom", "malibu", "mallard", "man", "manageme", "manager", "manahil", "manbat", "manchester", "mandy", "manfred", "mangesh", "mangue", "mani", "manish", "manohar", "manoj", "manon", "mansetmanis", "manson", "mantra", "manuel", "many", "maplesyrup", "mara", "marathon", "marc", "marcel", "marcella", "marcelle", "marcellin", "march", "marci", "marcia", "marcio", "marco", "marcus", "marcy", "mardi", "marek", "margalit", "margaret", "margarid", "margarit", "margaux", "marge", "margeaux", "margie", "margo", "marguerite", "maria", "maria1", "mariah", "mariah1", "marian", "marianne", "marie", "marie-madeleine", "marielle", "marietta", "mariette", "marilyn", "marina", "marine", "mariner", "marino", "mario", "marion", "mariposa", "maritial", "marius", "marjory", "mark", "mark1", "market", "marko", "markus", "marlboro", "marlena", "marlene", "marley", "marni", "marriage", "marrucci", "mars", "marshal", "marshall", "mart", "martha", "marthe", "marti", "martial", "martin", "martin1", "martine", "martinez", "martinien", "martiniq", "marty", "marvin", "mary", "maryam", "maryann", "maryjane", "marzec", "masahiro", "mase", "mash4077", "masoud", "mass", "masscomp", "master", "master1", "masters", "masuhiro", "math", "math-cs", "mathilde", "matilda", "matrix", "matt", "matther", "matthew", "matthias", "matthieu", "matti1", "mattingly", "maude", "maui", "maureen", "maurice", "mauricio", "mauro", "maveric", "maverick", "max", "maxime", "maxine", "maxmax", "maxwell", "maxwellsmart", "may", "mayday", "mazda1", "mazin", "me", "meagan", "meat", "meatcleaver", "meatloaf", "meatwagon", "mech", "mechanic", "medard", "media", "medical", "meekie", "mega", "megadeth", "megan", "meggie", "meister", "mel", "melaine", "melanie", "melina", "melinda", "melisa", "melissa", "mella", "mellon", "melody", "meltin", "member", "membership", "memory", "memphis", "men", "mendel", "mensuck", "meow", "mercedes", "mercer", "mercredi", "mercure", "mercury", "merde", "meres", "merlin", "merlot", "mermaid", "merrell", "merrill", "merrychristmas", "metal", "metallic", "metro", "mets", "mexico", "mgr", "miami", "michael", "michael.", "michael1", "michal", "michel", "michel1", "michele", "michell", "michelle", "michigan", "michou", "michy", "mickel", "mickey", "mickey1", "micky", "micro", "microsoft", "midnight", "midori", "midvale", "midway", "might", "miguel", "mihail", "mihran", "mikael", "mike", "mike1", "mikey", "miki", "mikko", "milano", "mildred", "miles", "milind", "milk", "millard", "millenium", "miller", "millicen", "millie", "million", "milo", "milton", "mimi", "mindy", "mine", "mined", "minerva", "ming", "minghe", "minh", "minimum", "minnie", "minot", "minou", "minsky", "minye", "miracle", "mirage", "miranda", "miriam", "mirror", "misanthrope", "misha", "mishka", "misogynist", "mission", "missirli", "missy", "misty", "mit", "mitch", "mitchell", "mittens", "mmm", "mmmm", "mmmmm", "mmmmmm", "mmmmmmm", "mmmmmmmm", "mmouse", "mnb", "mnbv", "mnbvc", "mnbvcx", "mnbvcxz", "mobile", "mobydick", "models", "modem", "modeste", "modula", "mogens", "mogul", "moguls", "mohamed", "mohammad", "mohammed", "mohan", "moises", "moishe", "moja", "mojo", "molly", "molly1", "molson", "molsongolden", "mom", "monday", "monet", "money", "money1", "moni", "monica", "monika", "monique", "monitor", "monkey", "monkey1", "monopoly", "monroe", "monster", "mont", "montana", "montana3", "month", "montreal", "montrose", "monty", "moocow", "mookie", "moomoo", "moon", "moonbeam", "moonpie", "moore", "moorhty", "moose", "moosehea", "moosehead", "mopar", "mora", "more", "morecats", "morgan", "morley", "moroni", "morpheus", "morris", "mort", "mortimer", "morts", "mose", "moshe", "mother", "motor", "motorola", "motown", "mou", "mountain", "mouse", "mouse1", "mousemat", "mousumi", "movies", "mowgli", "mozart", "mr.", "mr.roger", "mrcharlie", "mrgoodbar", "mrwonderful", "mt.xinu", "mtichell", "mtxinu", "muad-dib", "muaddib", "muamadin", "much", "muffin", "muh", "muhammad", "mukesh", "mukund", "mulder", "mulder1", "munaish", "munchkin", "mundeep", "murphy", "murray", "muscle", "music", "musicbox", "musim", "mustang", "mustang1", "mutant", "my", "mycroftxxx", "mycroftyyy", "mypasswd", "mypassword", "myra", "myron", "myrtle", "myself", "mysmut", "myung", "myung-yu", "nabil", "nadege", "nader", "nadia", "nadine", "naftaly", "nagel", "naissance", "nakamichi", "nalini", "nan", "nancy", "nanette", "naomi", "naoto", "napoleon", "narciso", "narcisse", "narendra", "nasa", "nascar", "nasty", "nat", "natacha", "natalia", "natalie", "nataraja", "natasha", "nathalie", "nathan", "nathanae", "nathanie", "nation", "national", "nationale", "nativite", "nautica", "naveen", "navette", "ncar", "ncc1701", "ncc1701d", "ncc1701e", "ncr", "ne1410s", "ne1469", "ne14a69", "neal", "nearmiss", "nebraska", "neckrub", "ned", "neenie", "neil", "neko", "nellie", "nelson", "nemesis", "nena", "nepenthe", "nepenthes", "neptune", "nermal", "nesbit", "nesbitt", "ness", "nestle", "nestor", "net", "netlinks", "netmgr", "netscape", "netware", "network", "networks", "neutrino", "nevada", "never", "neville", "new", "newaccount", "newblood", "newcourt", "newkid", "newkids", "newlife", "newpass", "news", "newton", "newuser", "newusers", "newworld", "newyork", "newyork1", "next", "nexus6", "nghi", "ngoc", "nguyen", "nicarao", "nicholas", "nichole", "nick", "nicklaus", "nicolas", "nicole", "niel", "nigel", "nigger", "nightmare", "nightshadow", "nightwal", "nightwind", "nihaoma", "nike", "nikhil", "niki", "nikita", "nikki", "nikolaos", "nilson", "nimh", "nimrod", "nina", "niners", "nino", "ninon", "nintendo", "nirvana", "nirvana1", "nissan", "nisse", "nita", "nite", "nnn", "nnnn", "nnnnn", "nnnnnn", "nnnnnnn", "nnnnnnnn", "no", "noam", "nobody", "nobuhiko", "nobuko", "noel", "nofun", "nokia", "nolan", "nomore", "nondet", "none", "none1", "nopass", "nora", "norbert", "noreen", "norene", "noriko", "norma", "normal", "norman", "northwest", "norton", "nosecret", "noshir", "not", "notebook", "notes", "notgay", "nothing", "notre", "notrespass", "notta1", "notused", "nouveau", "novacancy", "novell", "november", "novembre", "now", "noway", "noxious", "nroff", "nss", "nuclear", "nugget", "nuke", "nukem", "null", "number", "number1", "number9", "numberone", "numbers", "nurse", "nursie", "nut", "nutmeg", "nutrition", "nuucp", "nyquist", "oa", "oatmeal", "oaxaca", "obi", "obi kenobi", "obi wan kenobi", "obi-wan", "obiwan", "obsession", "ocean", "oceanography", "ocelot", "octavia", "october", "octobre", "odette", "odile", "odilon", "of", "off", "office", "ohshit", "ohwell", "oicu812", "oil", "oivind", "ojrind", "old", "oldlady", "oldpussy", "olin", "olive", "oliver", "olivetti", "olivia", "olivier", "ollie", "olsen", "omead", "omega", "on", "once", "one", "onionrings", "online", "only", "onstad", "ooo", "oooo", "ooooo", "oooooo", "ooooooo", "oooooooo", "oops", "open", "openbar", "opendesktop", "opendoor", "opensaysme", "opensesame", "openup", "oper", "opera", "operator", "opus", "or", "oracle", "orange", "orangeline", "oranges", "orca", "orchid", "oregon", "oreo", "orgasm", "orion", "orlando", "orville", "orwell", "oscar", "osiris", "osulliva", "oswald", "othar", "other", "otter", "otto", "ou812", "our", "oussama", "out", "outlaw", "outtolunch", "over", "overkill", "overthrow", "overtime", "owen", "own", "owns", "oxford", "ozzie", "ozzy", "paagal", "pacers", "pacific", "pacifique", "packard", "packer", "packers", "packrat", "pad", "padaaa", "paddy", "padma", "padoue", "page", "paige", "painless", "paint", "painter", "pakistan", "paladin", "pallab", "palmer", "paloma", "pam", "pamela", "pampers", "pancake", "panda", "pandora", "panic", "pantera", "panther", "panties", "papa", "paper", "papers", "papiers", "pappas", "paques", "paradigm", "parallel", "paranoia", "paraskev", "parfait", "paris", "park", "parker", "parkins", "parola", "parrot", "part", "partner", "parton", "parviz", "pascal", "pass", "passion", "passwd", "passwor", "password", "password1", "passwordlookhere", "pat", "patches", "patel", "paterne", "patrice", "patricia", "patrick", "patriots", "patsy", "patterson", "patti", "patton", "patty", "paul", "paula", "paule", "paulin", "pauline", "pavel", "pawan", "payman", "payment", "payton", "pcat", "pcxt", "peace", "peach", "peaches", "peanut", "peanutbutter", "peanuts", "pearl", "pearljam", "pebbles", "peche", "pecheur", "pecheurs", "pedro", "pedro1", "peebles", "peewee", "pegasus", "peggy", "peh", "pekka", "pelagie", "pencil", "penelope", "penguin", "penis", "penny", "pentecote", "pentium", "pentti", "people", "peoria", "pepper", "pepsi", "peraka", "percolate", "percy", "peres", "perez", "perfect", "performa", "perry", "persimmon", "person", "persona", "pervert", "pete", "peter", "peter1", "peterk", "peterpan", "peterson", "petey", "petunia", "peugeot", "peur", "pgonderin", "pham", "phantom", "phialpha", "phil", "philip", "philippe", "philips", "phillip", "phillips", "phish", "phishy", "phoenix", "phoenix1", "phone", "photo", "phrack", "phreak", "phrick", "phyllis", "piano", "piano1", "pianoman", "pianos", "picard", "picasso", "pick", "pickle", "picture", "pierce", "pierre", "pieter", "pigeon", "piglet", "pimp", "pin", "ping", "pink", "pinkfloy", "pinkfloyd", "pioneer", "piotr", "pipeline", "pipeorgan", "piper1", "pirate", "pirie", "pisces", "pit", "pizza", "plaintruth", "plane", "planet", "plato", "play", "playboy", "player", "players", "playground", "please", "plier", "plover", "plugh", "plumbrandy", "plus", "pluto", "pluton", "plymouth", "pmc", "pocus", "poetic", "poetry", "poh", "poir", "poire", "poisson", "poissons", "poiu", "poiuy", "poiuyt", "poiuytre", "polar", "polarbear", "polaris", "pole", "police", "politics", "polly", "polo", "polynomial", "pomme", "pondering", "pontiac", "pooh", "poohbear", "pookey", "pookie", "pookie1", "popcorn", "pope", "popeye", "poppy", "porc", "pork", "porky", "porn", "pornbay", "pornman", "porno", "pornography", "porsche", "porsche9", "porsche911", "porsche914", "porter", "portland", "portnoy", "postel", "poster", "powell", "power", "powertool", "ppp", "pppp", "ppppp", "pppppp", "ppppppp", "pppppppp", "prabhaka", "prabhu", "prabir", "pradeep", "praise", "pranab", "prasad", "prashant", "pratap", "pratt", "pravin", "prayer", "precious", "predator", "prelude", "premier", "presence", "present", "president", "presto", "preston", "pretty", "prettyface", "prevision", "price", "primus", "prince", "princess", "princeton", "print", "printemps", "printer", "printing", "prisca", "priv", "private", "privs", "producers", "prof", "prof.", "professor", "profile", "program", "promethe", "prometheus", "pronghorn", "property", "prosper", "protect", "protel", "protozoa", "provider", "prudence", "psalms", "psycho", "pub", "public", "pubus", "puckett", "puddin", "pull", "pulsar", "pumkinpie", "pumpkin", "puneet", "punkin", "pup", "puppet", "puppies", "puppy", "puppy123", "purnendu", "purple", "pussy", "pussy1", "put", "pyramid", "pyro", "python", "q1w2e3", "qed", "qian", "qinsong", "qqq", "qqq111", "qqqq", "qqqqq", "qqqqqq", "qqqqqqq", "qqqqqqqq", "quality", "quebec", "queen", "queenie", "quentin", "quest", "quoc", "qwaszx", "qwer", "qwert", "qwerty", "qwerty12", "qwertyu", "qwertyui", "r0ger", "r2d2", "rabbit", "rabbit1", "racer", "racerx", "rachel", "rachelle", "rachmaninoff", "racing", "racoon", "radar", "radha", "radio", "rafael", "raffi", "rafiki", "raghav", "raghavan", "raghu", "ragunath", "raid", "raider", "raiders", "raidersofthelostark", "raimund", "rain", "rainbow", "raindrop", "raissa", "raistlin", "raj", "raja", "rajadasa", "rajeeb", "rajeev", "rajendra", "rajiv", "rakesh", "raleigh", "ralph", "ram", "ramachan", "ramana", "ramani", "ramarao", "rambo", "rambo1", "rameaux", "ramesh", "ramirez", "ramon", "ran", "rancid", "randal", "randall", "randolph", "random", "randy", "randy1", "ranger", "rangers", "ranjan", "raoul", "rap", "raptor", "raquel", "rascal", "rasta", "rasta1", "rastafarian", "rastaman", "rat", "ratio", "raven", "ravens", "ravi", "ray", "raymona", "raymond", "read", "reader", "reading", "reagan", "real", "realfriend", "reality", "really", "realthing", "realtime", "rebecca", "rebels", "reboot", "red", "redbaron", "redbrick", "redcloud", "reddog", "redfish", "redline", "redman", "redrum", "redskin", "redskins", "redwing", "redwood", "reebok", "reed", "reefer", "reg", "reggae", "reggie", "regina", "reginald", "regional", "regis", "reine", "reliant", "remember", "remi", "remote", "remy", "renaud", "renault", "rene", "renee", "renegade", "rengaraj", "replicant", "repoman", "reponse", "reptile", "republic", "request", "requin", "rescue", "research", "reseau", "retard", "revolution", "rex", "reynolds", "reza", "reznor", "rfs", "rhett", "rhino", "rhjrjlbk", "rhona", "rhonda", "riacs", "ribs", "ricardo", "ricardo1", "riccardo", "rich", "richard", "richard1", "richards", "richardson", "richmond", "rick", "ricki", "ricky", "riddle", "ride", "riff-raff", "riffraf", "riffraff", "rifraf", "rifraff", "right", "riley", "ringo", "ripper", "ripple", "risc", "rit", "rita", "river", "rivera", "rivi", "rje", "road", "road warrior", "roadrunner", "roadwarrior", "rob", "robbie", "robby", "robert", "robert1", "roberta", "roberto", "roberts", "robin", "robinhoo", "robinhood", "robinson", "robley", "robocop", "robot", "robotech", "robotics", "robyn", "roche", "rochelle", "rochester", "rock", "rocket", "rocket1", "rockie", "rocknroll", "rockon", "rocky", "rocky horror", "rocky1", "rockyhorror", "rod", "rodent", "rodeo", "rodger", "rodman", "rodney", "rodolphe", "rodrigue", "rodriguez", "roger", "roger1", "rogers", "rohit", "roi", "rokny", "roland", "rolande", "rolex", "rolidex", "rollin", "rom", "romain", "roman", "romano", "romaric", "romeo", "rommel", "romuald", "romulan", "romulans", "romy", "ron", "ronak", "ronald", "ronen", "roni", "ronin", "ronitt", "rookie", "rooster", "root", "rootbeer", "roping", "rosa", "rosalie", "rose", "rosebud", "roseline", "rosemary", "roses", "rosie", "rosine", "ross", "rossigno", "roth", "rouge", "rough", "roulette", "round", "route66", "roxana", "roxy", "roy", "royal", "royals", "rrr", "rrrr", "rrrrr", "rrrrrr", "rrrrrrr", "rrrrrrrr", "rsm", "rti", "rtwo", "rtwodtwo", "ruben", "ruby", "rudolf", "rudy", "ruey", "rufus", "rugby", "rugger", "ruggieri", "ruknet", "rules", "run", "runner", "running", "ruoxin", "rush", "russ", "russel", "russell", "rusty", "ruth", "ruthie", "ruthless", "rux", "ruy", "ruye", "ryan", "ryohei", "ryota", "saab", "saab900", "saab9000", "saabturbo", "sabbath", "sabina", "sabine", "sabrina", "sacre", "sade", "sadie", "safaa", "safety", "safety1", "safwat", "sagittaire", "sai", "said", "saifalla", "saigon", "saikumar", "sailing", "sailor", "saint", "sainte", "sakti", "sakura", "sal", "sal9000", "salah", "salasana", "sales", "salle", "sally", "salmon", "salome", "salone", "salut", "sam", "samadams", "samantha", "samedi", "samiam", "samir", "sammie", "sammy", "sampath", "sample", "sampler", "sampson", "samsam", "samson", "samtaney", "samuel", "samurai", "sanchez", "sanders", "sanderson", "sandgorg", "sandi", "sandra", "sandrine", "sandsmmx", "sandy", "sanfran", "sanfrancisco", "sang", "sangbang", "sango", "sanh", "sani", "sanjay", "sanjeev", "sanjose", "sanjose1", "santa", "santiago", "santisuk", "santo", "saphire", "sapphire", "sara", "sarah", "sarah1", "saroj", "sasha", "sashi", "saskia", "sassy", "satori", "saturday", "saturn", "saturn5", "saturne", "saturnin", "saul", "sauvignon", "savage", "save", "sawedoff", "saxon", "say", "sbdc", "scamper", "scarecrow", "scarlet", "scarlett", "scheme", "schnapps", "school", "schroede", "science", "sciubba", "scooby", "scoobydoo", "scooter", "scooter1", "scorpio", "scorpion", "scot", "scotch", "scott", "scott1", "scottie", "scotty", "scout", "scouts", "scream", "scrofula", "scrooge", "scruffy", "scuba", "scuba1", "scumbag", "sdfghjkl", "sea", "seabreeze", "sean", "search", "seattle", "sebastien", "sechang", "secret", "secret3", "security", "see", "seeker", "seeme", "seho", "sei", "seigneur", "seive", "sekhar", "selftime", "semperfi", "sendit", "senior", "sensor", "seonghoo", "seoul", "septembe", "september", "septembre", "sequent", "serena", "serenity", "serge", "sergei", "sergey", "server", "service", "services", "sesame", "sesamestreet", "seth", "setup", "seung", "seunghyu", "seungku", "sevak", "seven", "seven7", "severin", "sex", "sexfiend", "sexxxme", "sexy", "sexyteen", "seymour", "shadow", "shadow1", "shadows", "shadyside", "shae", "shaggy", "shahrokh", "shakespeare", "shall", "shalom", "shamita", "shan", "shana", "shanghai", "shannan", "shannon", "shanny", "shantanu", "shanti", "shaolin", "sharad", "sharc", "shari", "shariyn", "shark", "sharks", "sharlene", "sharon", "sharra", "shashank", "shashi", "shasta", "shaun", "shaved", "shaven", "shaw", "shawn", "shayne", "shazam", "shazzam", "she", "sheba", "sheela", "sheena", "sheffield", "sheila", "shel", "shelby", "sheldon", "shelia", "shell", "shelley", "shelly", "shelter", "shen", "sheng", "shenglu", "shepherd", "sherif", "sherri", "sherrie", "sherry", "sheryl", "shi", "shiahn", "shidan", "shigenar", "shigeo", "shih", "shihming", "shimon", "shin", "shinobu", "ship", "shirin", "shirl", "shirley", "shit", "shit-head", "shitfaced", "shitforbrains", "shithead", "shiue", "shiva", "shivapra", "shivers", "shizoom", "shlee", "shlomo", "shoes", "shogun", "sholom", "shomita", "shooter", "shorty", "shotgun", "shout", "show", "shower", "showoff", "shrdlu", "shreeram", "shu", "shuang", "shuhui", "shun", "shutdown", "shuttle", "shyng", "sid", "sidarta", "sidekick", "sidharta", "sidney", "sidoine", "siemens", "sierra", "sigmachi", "signal", "signature", "sillywalk", "silver", "silvere", "silvia", "simba", "simba1", "simmons", "simon", "simple", "simply", "simpson", "simpsons", "simsim", "sina", "sinatra", "sing", "singer", "single", "siobahn", "siri", "sirius", "sister", "sit", "site", "siuping", "sivakuma", "six", "sixtynine", "sizenine", "skate", "skater", "skeeter", "skibum", "skidoo", "skiing", "skinny", "skip", "skipper", "skipper1", "skippy", "skull", "skunk", "skydive", "skyler", "skywalker", "slacker", "slayer", "sleazy", "sleep", "sleepy", "slick", "slider", "slimeball", "slinky", "slip", "slus", "slut", "small", "smallcock", "smallhips", "smalltalk", "smalltips", "smashed", "smashing", "smegma", "smile", "smile1", "smiles", "smiley", "smith", "smiths", "smitty", "smoke", "smokedham", "smokey", "smooch", "smother", "smurfy", "smut", "snafoo", "snafu", "snake", "snakes", "snappel", "snapper", "snapple", "snatch", "sneezy", "snell", "snicker", "snickers", "sniper", "snoop", "snoopdog", "snoopy", "snorkydorky", "snow", "snowbal", "snowball", "snowflake", "snowing", "snowman", "snowski", "snuffy", "so", "soap", "sober1", "soccer", "soccer1", "soccor", "socrate", "socrates", "sod", "sod off", "sodoff", "soft", "softball", "solange", "soleil", "soloman", "solomon", "soman", "somasama", "sombrero", "some", "somebody", "son", "sondra", "songmiao", "songnian", "sonia", "sonic", "sonics", "sonja", "sonny", "sony", "sonya", "soon", "soonman", "soowon", "sophia", "sophie", "sophomore", "sorel", "soroor", "sossina", "sotiris", "soua", "soumitra", "sound", "source", "sourire", "souris", "sourmilk", "souvenir", "space", "spaceman", "spain", "spam", "spanish", "spanky", "sparks", "sparky", "sparrow", "sparrows", "spartan", "spazz", "special", "specter", "spectre", "speech", "speed", "speedo", "speedy", "spence", "spencer", "sph", "sphynx", "spice", "spider", "spiderman", "spiff", "spiffy", "spike", "spike1", "spirit", "spiritu", "spiritusanctu", "spiro", "spiros", "spit", "spitfire", "splif", "spliff", "spock", "sponge", "spooky", "spooler", "spoon", "sports", "spot", "sprang", "spring", "springer", "sprite", "sprocket", "spud", "spunky", "spurs", "spyrogyra", "spyros", "squash", "squires", "squirt", "sridhar", "srimat", "srinivas", "sss", "ssss", "sssss", "ssssss", "sssssss", "ssssssss", "ssu", "stacey", "staci", "stacie", "stacy", "stainless", "stalker", "stamos", "stan", "stanislas", "stanley", "stanly", "stanton", "star", "star wars", "star69", "starbuck", "stargate", "starlight", "stars", "starship", "start", "starter", "startrek", "starwars", "stat", "states", "station", "status", "stealth", "steel", "steele", "steelers", "stefan", "stefano", "stella", "stemple", "steph", "steph1", "stephane", "stephani", "stephanie", "stephany", "stephen", "stephi", "stephon", "sterling", "stern93", "steve", "steve1", "steven", "steven1", "stevens", "stever", "stewart", "stickshift", "stiffdrink", "stiffprick", "stimpy", "stimulate", "sting", "sting1", "stingray", "stinky", "stivers", "stocks", "stone", "stop", "storage", "storem", "storm", "stormy", "strange", "stranger", "strangle", "strat", "stratford", "strato", "stratocaster", "strawber", "strawberry", "stretch", "strider", "strong", "sttng", "stu", "stuart", "stud", "student", "student2", "studfuck", "studio", "studly", "stuffed", "stuffedturkey", "stumpy", "stupid", "stuttgart", "su", "subgenius", "subhas", "subhdail", "subhednu", "subodh", "subscriber", "subway", "succes", "success", "suck", "sucker", "suckme", "suckrocks", "sucks", "sudeshna", "sudhakar", "sudhir", "sudir", "sue", "suesec", "sugar", "sugarbear", "sugih", "sukumar", "sultan", "sulu", "summer", "summit", "sumuinen", "sun", "sun-spot", "sunbird", "sundance", "sunday", "sunfire", "sunflowe", "sunflower", "sung", "sunil", "sunny", "sunny1", "sunnyvale", "sunrise", "sunset", "sunshin", "sunshine", "suntools", "sunwei", "super", "superfly", "superman", "superstage", "superstar", "superuser", "supervisor", "support", "supported", "supra", "suranet", "suresh", "surf", "surfer", "surfing", "susan", "susan1", "susanna", "susanne", "susha", "sushila", "susie", "sutton", "suvendu", "suvro", "suzanna", "suzanne", "suzie", "suzuki", "suzy", "sven", "sverige", "swamprat", "swane", "swanson", "swearer", "sweatshop", "sweden", "sweetie", "sweetnes", "sweetpea", "sweets", "sweety", "swim", "swimmer", "swimming", "swingset", "switzer", "swoosh", "swordfish", "syam", "sybase", "sybil", "sydney", "sylvain", "sylvere", "sylveste", "sylvestre", "sylvia", "sylvie", "symbol", "symmetry", "symult", "syphilis", "syphillis", "sys", "sys5", "sysadm", "sysadmin", "sysdiag", "sysdiags", "syslib", "sysmaint", "sysman", "sysmgr", "sysop", "system", "system5", "systemfive", "systemv", "systest", "sysv", "t-bone", "tab", "tabatha", "tacobell", "tadahiro", "tadlock", "taffy", "tai", "taiwan", "tajen", "tak", "taka", "takaji", "takashi", "take", "take5", "takefive", "takeiteasy", "takuji", "talon", "tam", "tamal", "tamara", "tamas", "tami", "tamie", "tammie", "tammy", "tamtam", "tandy", "tangerine", "tango", "tanguy", "tania", "tanju", "tanker", "tanner", "tanya", "tao", "tapani", "tapas", "tape", "tar", "tara", "tardis", "targas", "target", "tarheel", "tarragon", "tarzan", "tasha", "tata", "tatiana", "tatsuo", "tattoo", "tatum", "taureau", "taurus", "tayfur", "taylor", "tazdevil", "tazman", "tazmania", "tbird", "tchen", "tcp-ip", "tcp/ip", "tcpip", "teacher", "teachers", "teakettle", "teaparty", "tech", "technical", "techno", "ted", "teddi", "teddy", "teddy1", "teddybear", "tee", "teen", "teenager", "teeney", "teenfan", "teeny", "teflon", "telecast", "telecom", "telephone", "tell", "telnet", "temp", "temporal", "temptation", "temptress", "ten", "tennis", "tentation", "tequila", "teresa", "teri", "terminal", "terre", "terri", "terrill", "terry", "terry1", "test", "test1", "test123", "test2", "test3", "testcase", "tester", "testguy", "testi", "testing", "testtest", "testuser", "tetris", "tetsuo", "tex", "texas", "tgif", "thaddeus", "thailand", "thailande", "than", "thanasis", "thanh", "thank", "thankgod", "thankyou", "that", "thatcher", "thavy", "the", "theatre", "thebeef", "thebirds", "theboss", "thebutler", "thecle", "theend", "thegreatescape", "their", "thejudge", "theking", "thekingandi", "thelorax", "them", "theman", "themonkey", "then", "theodora", "theodore", "theophile", "thepenguin", "theproducers", "there", "therealthing", "theresa", "therese", "theriddler", "theron", "these", "they", "thiam", "thibault", "thibaut", "thickcock", "thickhead", "thickskin", "thierry", "thilaka", "think", "thinthighs", "this", "thisisit", "thoi", "thomas", "thompson", "thomson", "thorne", "thorsten", "those", "thrasher", "three", "thu", "thumper", "thunder", "thunderb", "thunderball", "thunderbird", "thunderdome", "thursday", "thuy", "thx1138", "tian", "tiffany", "tiger", "tiger2", "tigers", "tigger", "tightass", "tightcunt", "tightend", "tightfit", "tigre", "tijun", "tika", "till", "tim", "timber", "time", "timezone", "timothy", "tina", "ting", "tinker", "tinkerbell", "tinman", "tintin", "tireswing", "titanic", "tits", "tiw", "tiziano", "tjahjadi", "tnt", "to", "tobias", "toby", "tod", "today", "todd", "together", "toggle", "tohru", "tokyo", "tolkein", "tolkien", "tom", "tomate", "tomato", "tomcat", "tommy", "tone", "toni", "tony", "too", "tool", "toosilly", "tootsie", "topcat", "topgun", "topher", "topography", "tor", "torc", "tornado", "toronto", "torres", "torsten", "tortoise", "tortue", "toshiaki", "toshiba", "toshiter", "tosnow", "total", "tothedark", "toto1", "tototo", "toucan", "toufic", "tour", "toussaint", "tove", "toxic", "toy", "toyota", "toysrus", "traci", "tracie", "tractor", "tracy", "trailblazer", "trailer", "trails", "training", "tran", "transexual", "transfer", "transfiguration", "transit", "transmit", "transport", "trapdoor", "trapper", "trash", "trashcan", "travail", "travel", "travis", "tre", "treasure", "trebor", "tree", "trees", "trek", "trent", "trevor", "tri", "trial", "tribble", "tribbles", "tricia", "tricky", "trident", "trieu", "trina", "trish", "trisha", "tristan", "triton", "trivial", "trixie", "troff", "trojan", "trombone", "tron", "trophy", "trouble", "trout", "troy", "truc", "truck", "trucker", "trucks", "truefriend", "truelove", "truman", "trumpet", "trustno1", "try", "trya", "tse", "tsing", "tsing tao", "tsing-tao", "tsingtao", "tsung", "tsutomu", "ttt", "tttt", "ttttt", "tttttt", "ttttttt", "tttttttt", "tty", "tuan", "tuba", "tubas", "tucker", "tucson", "tuesday", "tula", "tull", "tunasalad", "tunasandwich", "tuomas", "turbo", "turbo2", "turner", "turnleft", "turnright", "turtle", "tuttle", "tuyen", "twat123", "tweetheart", "tweety", "twenex", "twila", "twins", "two", "tyler", "tyler1", "tzi", "tzila", "tzuwang", "uday", "udo", "uhn", "uhn-soon", "uhnsoon", "uli", "ulric", "ulrich", "ultimate", "umesh", "undead", "undergrad", "undergraduate", "unguessable", "unhappy", "unicorn", "unicos", "uniform", "unify", "unigrafix", "unique", "united", "unity", "unix", "unix-to-unix", "unix-tounix", "unixman", "unixsuck", "unixsux", "unknown", "unkown", "untung", "up", "upchuck", "upon", "upperclass", "upsilon", "upstill", "uptohere", "upyours", "uranus", "urbain", "urchin", "ursula", "us", "usa", "use", "usenet", "usenix", "user", "user1", "usermane", "username", "users", "util", "utility", "utopia", "utpal", "uucp", "uuu", "uuuu", "uuuuu", "uuuuuu", "uuuuuuu", "uuuuuuuu", "vacation", "vader", "vahe", "val", "valentin", "valentine", "valerie", "valerio", "valhalla", "valley", "vampire", "van", "vance", "vanessa", "vanilla", "varkey", "vasant", "vasanth", "vason", "vassilio", "vaughan", "vectrex", "vectrix", "veday", "vedder", "vee", "veljko", "velo", "velvet", "venceslas", "vendredi", "veneto", "venice", "venkat", "venkatad", "venkatar", "venkates", "venture", "venus", "ver", "verilog", "vermont", "vernon", "veronica", "veronique", "verseau", "vertige", "vertigo", "very", "veterans", "vette", "vianney", "vibeke", "vibhu", "vibrator", "vicesquad", "vicki", "vickie", "vicky", "victoire", "victor", "victor1", "victoria", "victorien", "victory", "video", "vierge", "vigyan", "vijay", "vijaya", "viking", "vikings", "vikram", "villa", "village", "vilma", "vinay", "vince", "vincent", "vincent1", "vinit", "vinitha", "vinod", "vinodh", "violet", "violin", "viper", "viper1", "virago", "virgil", "virgin", "virginbirth", "virginia", "virginie", "virginio", "virus", "visa", "visavis", "vishvjit", "vision", "visit", "visitation", "visitor", "vispi", "visual", "visvanat", "vittorio", "vivek", "vivian", "viviane", "vivien", "vjday", "vlad", "vladimir", "vlsi", "vm", "vms", "vmssucks", "vmsucks", "vojin", "volcano", "volley", "volleyb", "volvo", "voodoo", "vortex", "voyager", "vt100", "vt52", "vvv", "vvvv", "vvvvv", "vvvvvv", "vvvvvvv", "vvvvvvvv", "wade", "wai", "waiting", "walden", "waldo", "waleed", "walid", "walk", "walker", "walleye", "wally", "walter", "wan", "wanda", "wandojo", "wang", "wank", "wanker", "want", "wantmenow", "warcraft", "ward", "warez", "wargames", "warlock", "warm", "warmweather", "warner", "warren", "warrior", "warriors", "wars", "warthog", "was", "wash", "washington", "water", "water1", "waterloo", "watson", "waves", "way", "wayne", "wayne1", "we", "weasel", "web", "webetoys", "webhead", "webmaste", "webmaster", "webster", "wedge", "weenie", "weeny", "weezer", "wei", "weidong", "weiheng", "weinrich", "weiping", "welch", "welcome", "welcome1", "well", "welsher", "wen", "wendel", "wendell", "wendi", "wendy", "wendy1", "wengyik", "went", "were", "wes", "wesley", "western", "wet", "whale1", "whales", "what", "whatchamacallit", "whateve", "whatever", "whatnot", "whatsup", "whatsupdoc", "wheeling", "wheels", "when", "where", "whereisthebeef", "wheresthebeef", "whey", "which", "whisky", "whit", "white", "whiting", "whitney", "who", "whocares", "wholesale", "whoopie", "whoopy", "whore", "whoville", "why", "wibble", "wicked", "wilbur", "wildcat", "wildchild", "wilfried", "will", "willen", "william", "william1", "williams", "williamsburg", "willie", "willow", "willy", "wilma", "wilson", "win", "win95", "wind", "window", "windows", "windsurf", "winfred", "wing", "winner", "winnie", "winniethepooh", "winona", "winston", "winter", "wired", "wisconsin", "wisdom", "wish", "with", "witnessfortheprosecution", "wizard", "wizards", "wojtek", "wolf", "wolf1", "wolfgang", "wolfman", "wolverin", "wolverine", "wolves", "woman", "wombat", "wombat1", "women", "won", "wonder", "wonderboy", "wonderbread", "wong", "wonyun", "woobie", "woobin", "wooby", "wood", "woodland", "woodrow", "woodstoc", "woodwind", "woody", "woofwoof", "wooiyi", "wopperkenny", "word", "work", "world", "wormwood", "would", "wqsb", "wrangler", "wrestle", "wright", "write", "wun", "wuntsin", "www", "wwww", "wwwww", "wwwwww", "wwwwwww", "wwwwwwww", "wxyz", "wyldchyd", "wynne", "wyoming", "x-files", "x-men", "xanadu", "xanth", "xavier", "xaviere", "xcountry", "xfer", "xfiles", "xgeneration", "xi", "xiao", "xiaobo", "xiaogang", "xiaoli", "xiaomin", "xinghao", "xinu", "xmodem", "xray", "xue", "xueqing", "xwindows", "xxfreessxx", "xxpassxx", "xxsnowxx", "xxx", "xxx123", "xxxx", "xxxxx", "xxxxxx", "xxxxxxx", "xxxxxxxx", "xyz", "xyz123", "xyzzy", "y", "yabba", "yabba-dabba-doo", "yabbadabbadoo", "yaco", "yael", "yamaha", "yan", "yang", "yanjun", "yankee", "yankees", "yaomin", "yaser", "yee", "yello", "yellow", "yellowstone", "yeng", "yeon", "yeong", "yezi", "yiannis", "yigal", "yihua", "yin", "ying", "yingsha", "yingyang", "yishun", "yoda", "yodude", "yogesh", "yogibear", "yoichi", "yolanda", "yomama", "yon", "yonah", "yong", "yongdong", "yongho", "yonghwan", "yongsam", "yosemite", "yoshiaki", "yoshio", "you", "you'reok", "youareok", "youcef", "youhanse", "young", "your", "youreok", "yourself", "yuan", "yuehwern", "yugang", "yuh", "yuji", "yujiko", "yuka", "yukkei", "yukon", "yumi", "yumiko", "yung", "yuqian", "yuval", "yves", "yvette", "yvonne", "yyy", "yyyy", "yyyyy", "yyyyyy", "yyyyyyy", "yyyyyyyy", "zachary", "zack", "zap", "zapata", "zaphod", "zary", "zebra", "zebras", "zena69", "zener", "zenerdiode", "zenith", "zephyr", "zeppelin", "zepplin", "zeus", "zeusx", "zhaoqian", "zhaozhua", "zhengkun", "zhenyan", "zhi", "zhigang", "zhishun", "zhiwei", "zhixin", "zhongguo", "zhongmin", "ziggy", "zigzag", "zimmerman", "zinfandel", "zita", "ziv", "ziyou", "zmodem", "zoltan", "zombie", "zonda", "zoomer", "zoran", "zork", "zorkmid", "zoro", "zorro", "zulu", "zxc", "zxc123", "zxcv", "zxcvb", "zxcvbn", "zxcvbnm", "zz-top", "zztop", "zzz", "zzzz", "zzzzz", "zzzzzz", "zzzzzzz", "zzzzzzzz" ]
                list
